"""Accumulation policies will decide how control commands from multiple controllers (if present)
in the same control loop will be combined before sending them to the robot."""

from .cmd_accumulation_policies import CommandAccumulatorBase, SimpleCmdOverride



"""Provides simulation interfaces for robot extending and respecting the RobotInterface class.
These can be used in the control loop without needing a real robot.
"""

from .pybullet_robot import PybulletRobot



"""Module containing interface and definitions of User interfaces such as keyboard interface and
joystick interface that can be used in the control loop in place of true Global planners."""

from .ui_base import UIBase, DummyUI
from .keyboard_interface import KeyboardInterface



import logging
from typing import Any, Dict
from time import perf_counter
from inspect import currentframe, getframeinfo

# pylint:disable = W0221


class LoggingFormatter(logging.Formatter):
    """
    Custom logging formatter visually consistent with spdlog.
    """

    BOLD_RED: str = "\033[31;1m"
    BOLD_WHITE: str = "\033[37;1m"
    BOLD_YELLOW: str = "\033[33;1m"
    GREEN: str = "\033[32m"
    ON_RED: str = "\033[41m"
    RESET: str = "\033[0m"

    LEVEL_FORMAT: Dict[Any, str] = {
        logging.CRITICAL: f"[{ON_RED}{BOLD_WHITE}critical{RESET}]",
        logging.DEBUG: "[debug]",
        logging.ERROR: f"[{BOLD_RED}error{RESET}]",
        logging.INFO: f"[{GREEN}info{RESET}]",
        logging.WARNING: f"[{BOLD_YELLOW}warning{RESET}]",
    }

    def format(self, record):
        custom_format = (
            "[%(asctime)s] "
            + self.LEVEL_FORMAT.get(record.levelno, "[???]")
            + " %(message)s (%(filename)s:%(lineno)d)"
        )
        formatter = logging.Formatter(custom_format)
        return formatter.format(record)


logger = logging.getLogger()
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
handler.setFormatter(LoggingFormatter())
logger.addHandler(handler)
logging.basicConfig(level=logging.INFO)


class ThrottledLogger(logging.Logger):
    """Log data intermittently at specified frequency."""

    def __init__(self, name: str, rate: float, level: int | str = logging.INFO) -> None:
        super().__init__(name, level)
        _handler = logging.StreamHandler()
        _handler.setLevel(logging.DEBUG)
        _handler.setFormatter(LoggingFormatter())
        self.addHandler(_handler)
        self.__period = 1.0 / rate
        self.__next_tick = perf_counter() + self.__period

    def log(self, level, msg, *args, **kwargs):
        if self.__next_tick - perf_counter() <= 0.0:
            self.__next_tick = perf_counter() + self.__period
            super().log(level, msg, *args, **kwargs)


class _ThrottledLogging(logging.Logger):
    # automatically creates a logger and uses existing one for each line in a code.
    # Allows use directly as function call using
    # `throttled_logging.<logtype>("message", delay_time_in_sec)` where `throttled_logging`
    # is an alias of this class defined below.

    _ACTIVE_THROTTLED_LOGGERS = {}

    @staticmethod
    def _get_triggered_logger(suffix: str, rate: float) -> ThrottledLogger:
        info = getframeinfo(currentframe())
        logger_name = f"{info.filename}_{info.lineno}_{suffix}_logger"
        if logger_name not in _ThrottledLogging._ACTIVE_THROTTLED_LOGGERS:
            _ThrottledLogging._ACTIVE_THROTTLED_LOGGERS[logger_name] = ThrottledLogger(
                name=logger_name, rate=rate
            )
        return _ThrottledLogging._ACTIVE_THROTTLED_LOGGERS[logger_name]

    @staticmethod
    def info(msg, delay_sec, *args, **kwargs):
        _logger = _ThrottledLogging._get_triggered_logger("info", rate=1.0 / delay_sec)
        _logger.log(logging.INFO, msg, *args, **kwargs)

    @staticmethod
    def debug(msg, delay_sec, *args, **kwargs):
        _logger = _ThrottledLogging._get_triggered_logger("debug", rate=1.0 / delay_sec)
        _logger.log(logging.DEBUG, msg, *args, **kwargs)

    @staticmethod
    def warning(msg, delay_sec, *args, **kwargs):
        _logger = _ThrottledLogging._get_triggered_logger("warning", rate=1.0 / delay_sec)
        _logger.log(logging.WARNING, msg, *args, **kwargs)

    @staticmethod
    def error(msg, delay_sec, *args, **kwargs):
        _logger = _ThrottledLogging._get_triggered_logger("error", rate=1.0 / delay_sec)
        _logger.log(logging.ERROR, msg, *args, **kwargs)

    @staticmethod
    def critical(msg, delay_sec, *args, **kwargs):
        _logger = _ThrottledLogging._get_triggered_logger("critical", rate=1.0 / delay_sec)
        _logger.log(logging.CRITICAL, msg, *args, **kwargs)

    @staticmethod
    def fatal(msg, delay_sec, *args, **kwargs):
        _logger = _ThrottledLogging._get_triggered_logger("fatal", rate=1.0 / delay_sec)
        _logger.log(logging.FATAL, msg, *args, **kwargs)

    @staticmethod
    def log(level, delay_sec, msg, *args, **kwargs):
        _logger = _ThrottledLogging._get_triggered_logger("log", rate=1.0 / delay_sec)
        _logger.log(level, msg, *args, **kwargs)


throttled_logging = _ThrottledLogging  # pylint:disable=C0103
"""Log data intermittently once every t seconds. Allows use directly as function call using
`throttled_logging.<logtype>("message", delay_time_in_sec)`"""

__all__ = ["logging", "ThrottledLogger", "throttled_logging"]



from dataclasses import dataclass, field
from typing import Tuple, Mapping, TypeAlias, List
import copy
import numpy as np

from .tf_types import QuatType, Vector3D, Twist, Pose3D

QuatTrajType: TypeAlias = np.ndarray
"""Numpy array representating a trajectory of quaternions in format [[x,y,z,w]].
Shape: [t, 4], where t is the number of trajectory points.
"""

Vector3DTrajType: TypeAlias = np.ndarray
"""Numpy array representating a trajectory of 3D vectors in format [[x,y,z]].
Shape: [t, 3], where t is the number of trajectory points.
"""


@dataclass
class PointMotion:
    """Container to hold position, velocity, acceleration data of a point moving in 3D space."""

    position: np.ndarray = None
    """3D position of point."""
    velocity: np.ndarray = None
    """3D velocity of point."""
    acceleration: np.ndarray = None
    """3D acceleration of point."""

    def __post_init__(self):
        self.__do_checks()

    def __do_checks(self):
        if self.position is not None:
            assert (
                len(self.position) == 3
            ), f"Position dimension should be 3 for object of type {self.__class__.__name__}."
        if self.velocity is not None:
            assert (
                len(self.velocity) == 3
            ), f"Velocity dimension should be 3 for object of type {self.__class__.__name__}."
        if self.acceleration is not None:
            assert (
                len(self.acceleration) == 3
            ), f"Acceleration dimension should be 3 for object of type {self.__class__.__name__}."

    def __eq__(self, other: "PointMotion") -> bool:
        if self.position is not None and not np.all(self.position == other.position):
            return False
        if self.velocity is not None and not np.all(self.velocity == other.velocity):
            return False
        if self.acceleration is not None and np.all(self.acceleration == self.acceleration):
            return False
        return True

    def __setattr__(self, prop, val):
        if prop not in ["position", "velocity", "acceleration"]:
            raise NameError(
                f"Tried to assign value to illegal parameter {prop} in object ot type {self.__class__.__name__}."
            )
        if val is not None:
            val = np.array(val)
        super().__setattr__(prop, val)
        self.__do_checks()


@dataclass
class FrameMotion:
    """Container to hold pose, vel, acc data for an object/frame/vector moving in 3D space."""

    pose: Pose3D = None
    velocity: Twist = None
    acceleration: Twist = None

    def __eq__(self, other: "FrameMotion") -> bool:
        return (
            self.pose == other.pose
            and self.velocity == other.velocity
            and self.acceleration == other.acceleration
        )

    def __setattr__(self, prop, val):
        if prop not in ["pose", "velocity", "acceleration"]:
            raise NameError(
                f"Tried to assign value to illegal parameter {prop} in object ot type {self.__class__.__name__}."
            )
        if val is not None:
            if prop == "pose":
                assert isinstance(val, Pose3D)
            else:
                assert isinstance(val, Twist)
        super().__setattr__(prop, val)


def _get_closest_timestamp_idx(
    timestamps: np.ndarray,
    t: float,
    traj_start_time: float,
    traj_end_time: float,
    only_if_time_passed: bool,
    respect_traj_start_time: bool,
    respect_traj_end_time: bool,
) -> int:
    """NOTE: for internal use only"""
    if respect_traj_start_time and t < traj_start_time:
        raise ValueError(f"Trajectory does not have any value before timestamp={traj_start_time}.")
    if respect_traj_end_time and t > traj_end_time:
        raise ValueError(f"Trajectory does not have any value after timestamp={traj_end_time}.")
    # IF your array is sorted and is very large, this is a much faster search than np.where
    idx = np.searchsorted(timestamps, t, side="left")
    if idx > 0 and (
        idx == len(timestamps)
        # the check t < timestamps[idx] may have floating point precision errors
        # e.g., 1.99999999999999999999 < 2 may return False, because they could be equal
        or (only_if_time_passed and t < timestamps[idx])
        or np.fabs(t - timestamps[idx - 1]) <= np.fabs(t - timestamps[idx])
    ):
        return idx - 1
    else:
        return idx


@dataclass
class GeneralisedTrajectory:
    """Container for holding generalised N-D trajectories."""

    traj: np.ndarray
    """Sequence of ND positions in the trajectory. Shape [t,N]."""
    timestamps: np.ndarray = None
    """Optionally, each point in the trajectory can be timestamped with this [t] dim vector."""

    dim_names: List[str] = None
    """The name of each dimension of the N-D trajectory (e.g. joint names if this object contains
    joint trajectory)."""

    traj_start_time: float = field(default=None, init=False)
    """Start time for this trajectory. Only available if initialised with `timestamps`."""
    traj_end_time: float = field(default=None, init=False)
    """End time of this trajectory. Only available if initialised with `timestamps`."""

    def __post_init__(self):
        if self.timestamps is not None:
            self.traj_start_time = self.timestamps[0]
            self.traj_end_time = self.timestamps[-1]

    def __getitem__(self, idx: int) -> Tuple[float, np.ndarray]:
        """Subscripting this object will return the time, trajectory point at the specified index.

        Args:
            idx (int): The integer index.

        Returns:
            Tuple[float, np.ndarray]: Timestamp, trajectory point at the specified index.
        """
        assert (
            idx < self.traj.shape[0]
        ), f"Index {idx} out of bounds for trajectory of size {self.traj.shape[0]}."
        t = None if self.timestamps is None else self.timestamps[idx]
        return (t, self.traj[idx, :])

    def get_point_at_time(self, t: float) -> Tuple[int, float, np.ndarray]:
        """Get the index in list, timestamp and trajectory point from this trajectory given a
        valid timestamp.

        Args:
            t (float): The exact timestamp to retrieve the pose for.

        Raises:
            ValueError: Raised when the exact match for t is not found in this trajectory.

        Returns:
            Tuple[int, float, np.ndarray]: Index in list, timestamp, trajectory point at the specified time.
                NOTE: This method returns the redundant timestamp object as the first value in the
                returned tuple. This is to keep consistency with the `get_point_at_approx_time()` method.
                The value of timestamp will be the same as the input `t`, and can be ignored.
        """
        assert (
            self.timestamps is not None
        ), "This GeneralisedTrajectory object does not have `timestamps` attribute defined."
        try:
            idx = np.where(self.timestamps == t)[0][0]
        except IndexError as exc:
            raise ValueError(
                f"No exact match for timestamp {t} in trajectory. Try using `get_point_at_approx_time(t={t})`."
            ) from exc
        return idx, *self.__getitem__(idx)

    def get_point_at_approx_time(
        self,
        t: float,
        only_if_time_passed: bool = True,
        respect_traj_start_time: bool = True,
        respect_traj_end_time: bool = False,
    ) -> Tuple[int, float, np.ndarray]:
        """Get the index in list, timestamp and trajectory point at a time closest to the specified t.

        Args:
            t (float): The desired timestamp.
            only_if_time_passed (bool, optional): When this is enabled, the returned values are
                ensured to be before the next timestamp in the trajectory. This is useful
                when querying trajectory points in a loop with incrementing timestamp, and you
                want to be sure that the returned values are not from the future. This way a
                trajectory value is returned only if the specified time is >= its corresponding
                timestamp and before the next timestamp. E.g. this way if a planner/controller,
                uses this method to query a trajectory at a time, it does not return a value from
                future. Defaults to True.
            respect_traj_start_time (bool, optional): If set to true, raises ValueError if queried
                time is before self.traj_start_time. Defaults to True.
            respect_traj_end_time (bool, optional): If set to true, raises ValueError if queried
                time is after self.traj_end_time. Defaults to False.

        Returns:
            Tuple[int, float, np.ndarray]: Index in list, timestamp, trajectory point from this object.
        """

        assert (
            self.timestamps is not None
        ), "This GeneralisedTrajectory does not have `timestamps` attribute defined."
        idx = _get_closest_timestamp_idx(
            timestamps=self.timestamps,
            t=t,
            traj_start_time=self.traj_start_time,
            traj_end_time=self.traj_end_time,
            only_if_time_passed=only_if_time_passed,
            respect_traj_start_time=respect_traj_start_time,
            respect_traj_end_time=respect_traj_end_time,
        )
        return idx, *self.__getitem__(idx)


@dataclass
class Trajectory3D:
    """Container for holding 3D pose trajectories.

    NOTE: When creating this object, use `validate_trajectory=True` to assert trajectory validity.
        Defaults to True. Disable this to speed up construction.
    """

    position_traj: Vector3DTrajType
    """Sequence of 3D positions in the trajectory. Shape [t,3]."""
    orientation_traj: QuatTrajType
    """Sequence of quaternion orientations in the trajectory. Shape [t,4]."""
    timestamps: np.ndarray = None
    """Optionally, each point in the trajectory can be timestamped with this [t] dim vector."""

    frame_name: str = ""
    """The name of the frame whose trajectory is defined in this object."""

    validate_trajectory: bool = field(default=True, init=True, repr=False)
    """If this is enabled, trajectory validity will be checked after construction. Disable this for
    quicker initialisations. Defaults to True."""

    traj_start_time: float = field(default=None, init=False)
    """Start time for this trajectory. Only available if initialised with `timestamps`."""
    traj_end_time: float = field(default=None, init=False)
    """End time of this trajectory. Only available if initialised with `timestamps`."""

    def __post_init__(self):
        if self.validate_trajectory:
            # check if position traj and orientation traj lengths match
            assert (
                self.position_traj.shape[0] == self.orientation_traj.shape[0]
            ), f"Invalid trajectory for frame {self.frame_name}.."

            if self.timestamps is not None:
                # check if timestamps length makes sense
                assert (
                    self.timestamps.shape[0] == self.position_traj.shape[0]
                ), f"Timestamps array size ({self.timestamps.shape[0]}) does not match trajectory size ({self.position_traj.shape[0]}) for frame {self.frame_name}."
                # check if timestamps increase monotonically
                assert np.all(
                    self.timestamps == np.sort(self.timestamps)
                ), "Timestamps should be strictly increasing."

            # check position trajectory dimension
            assert (
                len(self.position_traj.shape) == 2 and self.position_traj.shape[1] == 3
            ), f"Position trajectory should be of shape [t,3] for frame {self.frame_name}."

            # check quaternion trajectory dimension
            assert (
                len(self.orientation_traj.shape) == 2 and self.orientation_traj.shape[1] == 4
            ), f"Orientation quaternions trajectory should be of shape [t,4] for frame {self.frame_name}."

            # check quaternion validity for each quaternion
            for nq in range(self.orientation_traj.shape[0]):
                assert np.isclose(
                    np.linalg.norm(self.orientation_traj[nq, :]), 1.0
                ), f"Quaternion at index {nq} in trajectory is not normalised for frame {self.frame_name}. Culprit: {self.orientation_traj[nq,:]}."
        if self.timestamps is not None:
            self.traj_start_time = self.timestamps[0]
            self.traj_end_time = self.timestamps[-1]

    def __getitem__(self, idx: int) -> Tuple[float, Vector3D, QuatType]:
        """Subscripting this object will return the time, position, orientation values at the specified index.

        Args:
            idx (int): The integer index.

        Returns:
            Tuple[float, Vector3D, QuatType]: Timestamp, position, orientation at the specified index.
        """
        assert (
            idx < self.position_traj.shape[0]
        ), f"Index {idx} out of bounds for trajectory of size {self.position_traj.shape[0]}."
        t = None if self.timestamps is None else self.timestamps[idx]
        return (
            t,
            self.position_traj[idx, :],
            self.orientation_traj[idx, :],
        )

    def get_pose_at_time(self, t: float) -> Tuple[int, float, Vector3D, QuatType]:
        """Get the index in list, timestampe, position and orientation values from this trajectory
        given a valid timestamp.

        Args:
            t (float): The exact timestamp to retrieve the pose for.

        Raises:
            ValueError: Raised when the exact match for t is not found in this trajectory.

        Returns:
            Tuple[int, float, Vector3D, QuatType]: Index in list, timestamp, position and quaternion
                values at the specified time.
                NOTE: This method returns the redundant timestamp object as the first value in the
                returned tuple. This is to keep consistency with the `get_pose_at_approx_time()`
                method. The value of timestamp will be the same as the input `t`, and can be
                ignored.
        """
        assert (
            self.timestamps is not None
        ), "This Trajectory3D does not have `timestamps` attribute defined."
        try:
            idx = np.where(self.timestamps == t)[0][0]
        except IndexError as exc:
            raise ValueError(
                f"No exact match for timestamp {t} in trajectory. Try using `get_pose_at_approx_time(t={t})`."
            ) from exc
        return idx, *self.__getitem__(idx)

    def get_pose_at_approx_time(
        self,
        t: float,
        only_if_time_passed: bool = True,
        respect_traj_start_time: bool = True,
        respect_traj_end_time: bool = False,
    ) -> Tuple[int, float, Vector3D, QuatType]:
        """Get the index, timestamp, position and orientation value at a time closest to the
        specified t.

        Args:
            t (float): The desired timestamp.
            only_if_time_passed (bool, optional): When this is enabled, the returned values are
                ensured to be before the next timestamp in the trajectory. This is useful
                when querying trajectory points in a loop with incrementing timestamp, and you
                want to be sure that the returned values are not from the future. This way a
                pose value is returned only if the specified time is >= its corresponding
                timestamp and before the next timestamp. E.g. this way if a planner/controller,
                uses this method to query a pose at a time, it does not return a value from the
                future. Defaults to True.
            respect_traj_start_time (bool, optional): If set to true, raises ValueError if queried
                time is before self.traj_start_time. Defaults to True.
            respect_traj_end_time (bool, optional): If set to true, raises ValueError if queried
                time is after self.traj_end_time. Defaults to False.

        Returns:
            Tuple[int, float, Vector3D, QuatType]: Index in list, closest timestamp, position,
                and quaternion from this Trajectory3D object.
        """

        assert (
            self.timestamps is not None
        ), "This Trajectory3D does not have `timestamps` attribute defined."
        idx = _get_closest_timestamp_idx(
            timestamps=self.timestamps,
            t=t,
            traj_start_time=self.traj_start_time,
            traj_end_time=self.traj_end_time,
            only_if_time_passed=only_if_time_passed,
            respect_traj_start_time=respect_traj_start_time,
            respect_traj_end_time=respect_traj_end_time,
        )
        return idx, *self.__getitem__(idx)


@dataclass
class MultiFrameTrajectory3D:
    """Container for holding synchronised trajectories for multiple frames.

    This object is also a single-use iterator, and can therefore be used in a loop, or using
    the `next(mf_traj)` to get the next `__getitem__` call. However, being a single-use
    iterator, it cannot be used again as an iterable once it hits `StopIteration`.

    WARNING
    -------
    Avoid construction in control loop. Can be heavy.

    Args:
        trajectories (List[Trajectory3D]): List of Trajectory 3D object holding pose trajectories
            for different frames.
        timestamps (np.ndarray): Timestamps associated with the waypoints in the trajectories.
            Each point in the trajectory (for all frames) is timestamped with this [t] dim vector.
            NOTE: all trajectories will be assumed to be using the same timestamps defined here.
    """

    trajectories: List[Trajectory3D]
    """List of Trajectory 3D object holding pose trajectories for different frames."""

    timestamps: np.ndarray
    """Each point in the trajectory (for all frames) is timestamped with this [t] dim vector.
    NOTE: all trajectories will be assumed to be using the same timestamps defined here.
    """

    traj_start_time: float = field(default=None, init=False)
    """Start time for this trajectory."""
    traj_end_time: float = field(default=None, init=False)
    """End time of this trajectory."""

    _curr_idx: int = field(repr=False, init=False)
    _max_idx: int = field(repr=False, init=False)

    @classmethod
    def createFromKeywordArgs(
        cls: "MultiFrameTrajectory3D",
        timestamps: List[float] | np.ndarray,
        **kwargs: Tuple[Vector3DTrajType, QuatTrajType],
    ) -> "MultiFrameTrajectory3D":
        """Classmethod to construct a MultiFrameTrajectory3D object using keywords for each frame'
        trajectory.

        Args:
            timestamps (List[float] | np.ndarray): List of timestamps to be used for all frames
            **kwargs (Tuple[Vector3DTrajType, QuatTrajType]): Each keyword argument should be the
                name of the frame mapping to their corresponding position and orientation
                trajectory. e.g.

        MultiFrameTrajectory3D.createFromKeywordArgs(timestamps = ts, ee_1 = [postraj_1, quattraj1],
            some_other_frame = [postraj_2, quattraj_2])

        Returns:
            MultiFrameTrajectory3D: A MultiFrameTrajectory3D object with the specified frame
            trajectories.
        """
        assert (
            len(kwargs) > 0
        ), "Trajectory for frames should be passed using the frame name as keyword."

        traj_list = []

        for frame_name, kwarg in kwargs.items():
            assert (
                len(kwarg) == 2
            ), "Each frame (key) should have as value a tuple: [list of positions, list of quaternions]."
            position_traj = kwarg[0]
            quat_traj = kwarg[1]

            traj = Trajectory3D(
                position_traj=position_traj,
                orientation_traj=quat_traj,
                timestamps=timestamps,
                frame_name=frame_name,
                validate_trajectory=True,
            )

            traj_list.append(copy.deepcopy(traj))

        return cls(trajectories=traj_list, timestamps=timestamps)

    def __post_init__(self):
        if not isinstance(self.timestamps, np.ndarray):
            self.timestamps = np.array(self.timestamps)

        assert len(self.timestamps.shape) == 1

        self.traj_start_time = self.timestamps[0]
        self.traj_end_time = self.timestamps[-1]

        # check trajectory length for each frame
        for traj in self.trajectories:
            assert (
                self.timestamps.shape[0]
                == traj.position_traj.shape[0]
                == traj.orientation_traj.shape[0]
            )
            # check if each trajectory has a valid frame name
            assert (
                traj.frame_name is not None and traj.frame_name != ""
            ), "Each trajectory should have a defined frame_name"

        self._curr_idx = -1
        self._max_idx = self.timestamps.shape[0]

    def __getitem__(self, idx: int) -> Tuple[float, Mapping[str, Tuple[Vector3D, QuatType]]]:
        """Subscripting this object will return the timestamp, as well pose per frame at the
        specified index.

        Args:
            idx (int): The integer index.

        Returns:
            Tuple[float, Mapping[str, Tuple[Vector3D, QuatType]]]: Timestamp, Dictionary with
                mapping from frame_name -> [position, quaternion].
        """
        assert (
            idx < self.timestamps.shape[0]
        ), f"Index {idx} out of bounds for trajectory of size {self.timestamps.shape[0]}."
        return (
            self.timestamps[idx],
            {
                traj.frame_name: [
                    traj.position_traj[idx, :],
                    traj.orientation_traj[idx, :],
                ]
                for traj in self.trajectories
            },  # TODO: would this be mutable?? # pylint: disable=W0511
        )

    def get_frame_poses_at_time(
        self, t: float
    ) -> Tuple[int, float, Mapping[str, Tuple[Vector3D, QuatType]]]:
        """Get the pose for each frame from this object given a valid timestamp.

        Args:
            t (float): The timestamp to retrieve the pose for.

        Raises:
            ValueError: Raised when the exact match for t is not found in this trajectory.

        Returns:
            Tuple[int, float, Mapping[str, Tuple[Vector3D, QuatType]]]: Index in list, Timestamp,
                Dictionary with mapping from frame_name -> [position, quaternion] at the specified
                timestamp.
                NOTE: This method returns the redundant timestamp object as the first value in the
                returned tuple. This is to keep consistency with the `get_pose_at_approx_time()`
                method. The value of timestamp will be the same as the input `t`, and can be
                ignored.
        """
        try:
            idx = np.where(self.timestamps == t)[0][0]
        except IndexError as exc:
            raise ValueError(
                f"No exact match for timestamp {t} in trajectory. Try using `get_pose_at_approx_time(t={t})`."
            ) from exc
        return idx, *self.__getitem__(idx)

    def get_frame_poses_at_approx_time(
        self,
        t: float,
        only_if_time_passed: bool = True,
        respect_traj_start_time: bool = True,
        respect_traj_end_time: bool = False,
    ) -> Tuple[int, float, Mapping[str, Tuple[Vector3D, QuatType]]]:
        """Get the pose for each frame at a time closest to the specified t.

        Args:
            t (float): The desired timestamp.
            only_if_time_passed (bool, optional): When this is enabled, the returned values are
                ensured to be before the next timestamp in the trajectory. This is useful
                when querying trajectory points in a loop with incrementing timestamp, and you
                want to be sure that the returned values are not from the future. This way a
                pose value is returned only if the specified time is >= its corresponding
                timestamp and before the next timestamp. E.g. this way if a planner/controller,
                uses this method to query a pose at a time, it does not return a value from the
                future. Defaults to True.
            respect_traj_start_time (bool, optional): If set to true, raises ValueError if queried
                time is before self.traj_start_time. Defaults to True.
            respect_traj_end_time (bool, optional): If set to true, raises ValueError if queried
                time is after self.traj_end_time. Defaults to False.

        Returns:
            Tuple[int, float, Mapping[str, Tuple[Vector3D, QuatType]]]: Index in list, Timestamp,
                Dictionary with mapping from frame_name -> [position, quaternion] that is
                closest to the specified time in this MultiFrameTrajectory3D object.
        """
        idx = _get_closest_timestamp_idx(
            timestamps=self.timestamps,
            t=t,
            traj_start_time=self.traj_start_time,
            traj_end_time=self.traj_end_time,
            only_if_time_passed=only_if_time_passed,
            respect_traj_start_time=respect_traj_start_time,
            respect_traj_end_time=respect_traj_end_time,
        )
        return idx, *self.__getitem__(idx)

    def __iter__(self):
        return self

    def __next__(self):
        self._curr_idx += 1
        if self._curr_idx >= self._max_idx:
            raise StopIteration
        return self.__getitem__(self._curr_idx)



"""Accumulation policies are functions that takes in commands from two controllers
and decides how to generate a single command to be sent to the robot.

NOTE: this should ideally be done at joint level, but we are simplifying by doing
accumulation at robot-level instead.
"""

from typing import List
from abc import ABC, abstractmethod
import copy

from ....core.types.robot_io import RobotCmd


class CommandAccumulatorBase(ABC):
    """Accumulation policies are functions that takes in commands from two controllers
    and decides how to generate a single command to be sent to the robot.

    NOTE: this should ideally be done at joint level, but we are simplifying by doing
    accumulation at robot-level instead."""

    @abstractmethod
    def accumulate(self, commands=List[RobotCmd]) -> RobotCmd:
        """Accumulate the specified commands into a single command to be sent to the robot.

        Args:
            commands (List[RobotCmd]): the robot commands from different controllers (in sequence).

        Returns:
            RobotCmd: The final control command (action) to be sent to the robot.

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.
        """
        raise NotImplementedError("This method has to be implemented in the child class")


class SimpleCmdOverride(CommandAccumulatorBase):
    """This accumulation policy simply overrides the command from all controllers with the latest
    one."""

    def accumulate(self, commands=List[RobotCmd]) -> RobotCmd:
        return commands[-1]


class CmdMuxer(CommandAccumulatorBase):
    """This policy combines the commands from all controllers/agents."""

    def __init__(self, allow_conflicting_interfaces: bool = False):
        self._cmd: RobotCmd = None
        self._allow_conflicts = allow_conflicting_interfaces

    def accumulate(self, commands=List[RobotCmd]) -> RobotCmd:
        # return super().accumulate(commands)
        if self._cmd is None:
            self._cmd = copy.deepcopy(commands[0])
            if len(commands) > 1:
                for cmd in commands[1:]:
                    self._cmd.extend(cmd, run_checks=True, override_existing=self._allow_conflicts)



from typing import Mapping
from contextlib import contextmanager
import os
import tempfile
import yourdfpy
import numpy as np
from scipy.spatial.transform import Rotation


def _convert_to_fixed(urdf: yourdfpy.URDF, joint_names_to_values: Mapping[str, float]):
    """Hack to modify urdf to fix specific joints to specified values.

    Args:
        urdf (yourdfpy.URDF): The yourdfpy urdf object to modify.
        joint_names_to_values (Mapping[str, float]): Dictionary containing name of joints that
            should be locked as keys, and the joint values to lock them at as values.
    """
    robot = urdf.robot
    for joint in robot.joints:
        if joint.name in joint_names_to_values.keys():
            # joints are modified by changing the origin of the
            # joint location wrt to the parent link
            new_local_transformation = np.eye(4)
            if joint.type == "prismatic":
                new_local_transformation[np.where(joint.axis == 1)[0], 3] = joint_names_to_values[
                    joint.name
                ]
            elif joint.type in ["revolute", "continuous"]:
                local_rpy = np.zeros(3)
                local_rpy[np.where(joint.axis == 1)[0]] = joint_names_to_values[joint.name]
                new_local_transformation[:3, :3] = Rotation.from_euler("xyz", local_rpy).as_matrix()
            joint.origin = joint.origin @ new_local_transformation
            joint.type = "fixed"


@contextmanager
def temp_path_urdf_with_joints_fixed(
    urdf_file_path: str, joint_names_to_values: Mapping[str, float]
):
    """Get the context of a temporary URDF file path after modifying the specified joints to be
    "fixed".

    This method only provides a context. This is useful for temporarily writing the URDF to a file
    for consumption by other programs.

    Args:
        urdf_file_path (str): Path to original urdf file.
        joint_names_to_values (Mapping[str, float]): Dictionary containing name of joints that
            should be locked as keys, and the joint values to lock them at as values.

    Yields:
        str: Path to new urdf file.

    Examples
    --------
    Use as a context manager:
        >>> with temp_path_urdf_with_joints_fixed(...) as path:
                load_file_from_path(path)  # do something with this urdf
    """
    urdf = yourdfpy.URDF.load(urdf_file_path)
    _convert_to_fixed(urdf=urdf, joint_names_to_values=joint_names_to_values)

    _, new_urdf_path = tempfile.mkstemp(suffix=".urdf")
    try:
        urdf.write_xml_file(new_urdf_path)
        yield new_urdf_path
    finally:
        os.remove(new_urdf_path)


def create_new_urdf_with_joints_fixed(
    urdf_file_path: str, joint_names_to_values: Mapping[str, float]
) -> str:
    """Creates a new urdf file with the specified joints set to "fixed".

    Args:
        urdf_file_path (str): Path to original urdf file.
        joint_names_to_values (Mapping[str, float]): Dictionary containing name of joints that
            should be locked as keys, and the joint values to lock them at as values.

    Returns:
        str: Path to new urdf file.
    """

    urdf = yourdfpy.URDF.load(urdf_file_path)
    _convert_to_fixed(urdf=urdf, joint_names_to_values=joint_names_to_values)

    new_urdf_path = urdf_file_path[:-5] + "_modified_fixed_joints.urdf"
    urdf.write_xml_file(new_urdf_path)

    return new_urdf_path



from typing import Tuple, Callable, TypeAlias, List
import pybullet as pb
import numpy as np

from ...utils.time_utils import RateTrigger, PythonPerfClock
from ...utils.math_utils import quat2rot

QuatType: TypeAlias = np.ndarray
"""Numpy array representating quaternion in format [x,y,z,w]"""
Vector3D: TypeAlias = np.ndarray
"""Numpy array representating 3D cartesian vector in format [x,y,z]"""


class PybulletGUIButton:
    """Create a button in Pybullet GUI, and check if it was pressed.
    Meant to be used in a loop."""

    def __init__(self, name: str, cid: int):
        """Create a button in Pybullet GUI, and check if it was pressed.
        Meant to be used in a loop.

            Args:
                name (str): Name on button.
                cid (int): The client id to connect to the physics simulator.
        """
        self.cid = cid
        self._id = pb.addUserDebugParameter(name, 1, 0, 1, physicsClientId=self.cid)
        self._val = 1

    def reset(self):
        """Not really useful anymore (legacy)"""
        self._val = 1

    def was_pressed(self) -> bool:
        """Checks if this button was pressed since the last time this method was called."""
        new_val = pb.readUserDebugParameter(self._id, physicsClientId=self.cid)
        triggered = new_val > self._val
        self._val = new_val
        return triggered

    def remove(self):
        """Remove the button from the gui."""
        try:
            pb.removeUserDebugItem(self._id, physicsClientId=self.cid)
        except pb.error:
            pass


class PybulletGUISlider:
    """Create a slider gui in pybullet."""

    def __init__(
        self,
        name: str,
        lower_lim: float,
        upper_lim: float,
        default_val: float,
        cid: int,
        mapping_function: Callable[[float], float] = lambda x: x,
    ):
        """Create a slider gui in pybullet.

        Args:
            name (str): Name for slider.
            lower_lim (float): The min value for slider.
            upper_lim (float): The max value for slider.
            default_val (float): Default starting value for slider.
            cid (int): The client id to connect to the physics simulator.
            mapping_function (Callable[[float], float], optional): Mapping to do for the
                value read from the slider (e.g. convert radian to degrees)
        """
        self.cid = cid
        self._id = pb.addUserDebugParameter(
            name,
            lower_lim,
            upper_lim,
            default_val,
            physicsClientId=self.cid,
        )
        assert callable(mapping_function)
        self._mapping_function = mapping_function
        self._last_recorded_val = None

    def get_value(self) -> float:
        """Read the value from the slider gui."""
        self._last_recorded_val = self._mapping_function(
            pb.readUserDebugParameter(self._id, physicsClientId=self.cid)
        )
        return self._last_recorded_val

    @property
    def last_read_value(self) -> float:
        return self._last_recorded_val

    def value_changed(self) -> bool:
        latest_val = self._last_recorded_val
        new_val = self.get_value()
        return new_val != latest_val

    def remove(self):
        """Remove the slider from gui."""
        try:
            pb.removeUserDebugItem(self._id, physicsClientId=self.cid)
        except pb.error:
            pass

    def set_value(self, value: float):
        raise NotImplementedError(
            "Pybullet slider values cannot be updated! Use TkInter sliders instead."
        )


class PybulletDebugFrameViz:
    """Draw a coordinate frame in pybullet as specified pose in the world."""

    def __init__(
        self,
        cid: int,
        position: Vector3D = None,
        orientation: QuatType = np.array([0, 0, 0, 1]),
        line_length: float = 0.2,
        line_width: float = 2,
        duration: float = 0.0,
        draw_at_init: bool = True,
    ):
        """Draw a coordinate frame in pybullet as specified pose in the world.

        Args:
            position (Vector3D): Position of coordinate frame in world.
            orientation (QuatType): Orientation quaternion of the frame in the world.
            cid (int): The physics server id to connect to.
            line_length (float, optional): Length of the lines in the world. Defaults to 0.2.
            line_width (float, optional): Line thickness. Defaults to 2.
            duration (float, optional): Lifetime for the frame. Defaults to 0.0 (infinite).
        """
        self._len = line_length
        self._lw = line_width
        self._duration = duration
        self._cid = cid

        self._line_ids = [-1, -1, -1]
        if draw_at_init:
            assert (
                position is not None and orientation is not None
            ), "Position and orientation for coordinate frame not provided."
            self._draw_lines(pos=position, ori=orientation)

    def _draw_lines(
        self,
        pos: Vector3D,
        ori: QuatType,
    ):
        rot_mat = quat2rot(ori).T
        toX = pos + self._len * rot_mat[0, :]
        toY = pos + self._len * rot_mat[1, :]
        toZ = pos + self._len * rot_mat[2, :]
        line_ids = self._line_ids
        self._line_ids[0] = pb.addUserDebugLine(
            pos,
            toX,
            [1, 0, 0],
            self._lw,
            self._duration,
            physicsClientId=self._cid,
            replaceItemUniqueId=line_ids[0],
        )
        self._line_ids[1] = pb.addUserDebugLine(
            pos,
            toY,
            [0, 1, 0],
            self._lw,
            self._duration,
            physicsClientId=self._cid,
            replaceItemUniqueId=line_ids[1],
        )
        self._line_ids[2] = pb.addUserDebugLine(
            pos,
            toZ,
            [0, 0, 1],
            self._lw,
            self._duration,
            physicsClientId=self._cid,
            replaceItemUniqueId=line_ids[2],
        )

    def update_frame_pose(self, position: Vector3D, orientation: QuatType):
        """Update the pose of an existing frame viz in pybullet.

        Args:
            position (Vector3D): Position of coordinate frame in world.
            orientation (QuatType): Orientation quaternion of the frame in the world.
        """
        self._draw_lines(pos=position, ori=orientation)

    def remove(self):
        """Remove the frame viz from gui."""

        try:
            for l_id in self._line_ids:
                pb.removeUserDebugItem(l_id, physicsClientId=self._cid)
        except pb.error:
            pass
        self._line_ids = [-1, -1, -1]


class PybulletText:
    """Add text to pybullet sim world gui."""

    def __init__(
        self,
        text_position: Tuple[float, float, float],
        default_text: str = "",
        text_color: Tuple[float, float, float] = (0, 0, 0),
        text_size: float = 3.0,
        lifetime: float = 0.0,
        cid: int = 0,
    ):
        """Add text to pybullet sim world gui.

        Args:
            text_position (Tuple[float, float, float]): 3D position to place the text in world
                coordinates.
            default_text (str, optional): Text to show. Defaults to "".
            text_color (Tuple[float, float, float], optional): RGB color code (range [0,1]).
                Defaults to (0, 0, 0).
            text_size (float, optional): Default text size. Defaults to 3.0.
            lifetime (float, optional): How long (in sec) the text should show. Defaults to 0.0
                (persistent).
            cid (int, optional): The client id to connect to the physics simulator. Defaults to 0.
        """
        self.cid = cid
        self._pos = text_position
        self._color = text_color
        self._size = text_size
        self._t = lifetime
        self._id = pb.addUserDebugText(
            default_text,
            self._pos,
            self._color,
            self._size,
            self._t,
            physicsClientId=self.cid,
        )

    def update_text(self, text: str):
        """Update the text string for this debug object.

        Args:
            text (str): The new text to replace with.
        """
        prev_id = self._id
        self._id = pb.addUserDebugText(
            str(text),
            self._pos,
            self._color,
            self._size,
            self._t,
            replaceItemUniqueId=prev_id,
            physicsClientId=self.cid,
        )

    def remove(self):
        """Remove text from pybullet."""
        try:
            pb.removeUserDebugItem(self._id, physicsClientId=self.cid)
        except pb.error:
            pass


class PybulletTextWithRateTrigger(PybulletText):
    """Add text to pybullet sim world gui, that will only change at the
    specified rate (even if `update_text` is called at a higher frequency).
    This is useful if this method is called in a high frequency loop such as
    a control loop, and text has to be rendered only at a lower frequency.
    """

    def __init__(
        self,
        text_position: Tuple[float, float, float],
        rate: float = 1,
        default_text: str = "",
        text_color: Tuple[float, float, float] = (0, 0, 0),
        text_size: float = 3.0,
        lifetime: float = 0.0,
        cid: int = 0,
    ):
        """Add text to pybullet sim world gui, that will only change at the
        specified rate (even if `update_text` is called at a higher frequency).
        This is useful if this method is called in a high frequency loop such as
        a control loop, and text has to be rendered only at a lower frequency.

        Args:
            text_position (Tuple[float, float, float]): 3D position to place the text in world
                coordinates.
            rate (float, optional): Update rate for text updating. Defaults to 1.
            default_text (str, optional): Text to show. Defaults to "".
            text_color (Tuple[float, float, float], optional): RGB color code (range [0,1]).
                Defaults to (0, 0, 0).
            text_size (float, optional): Default text size. Defaults to 3.0.
            lifetime (float, optional): How long (in sec) the text should show. Defaults to 0.0
                (persistent).
            cid (int, optional): The client id to connect to the physics simulator. Defaults to 0.
        """
        self._trigger = RateTrigger(rate=rate, clock=PythonPerfClock())
        super().__init__(
            text_position=text_position,
            default_text=default_text,
            text_color=text_color,
            text_size=text_size,
            lifetime=lifetime,
            cid=cid,
        )

    def update_text(self, text: str):
        """Update the text string for this debug object. Will only be triggered if the specified
        rate is satisified (lower bound).

        Args:
            text (str): The new text to replace with.
        """
        if self._trigger():
            return super().update_text(text)


class PybulletDebugPoints:
    """Add 3D points to pybullet sim world gui."""

    def __init__(
        self,
        point_positions: List[Vector3D] = None,
        point_size: float = 1.0,
        rgb: Tuple[float, float, float] = (1, 0, 0),
        lifetime: float = 0.0,
        cid: int = 0,
    ):
        """Add 3D points to pybullet sim world gui.

        Args:
            point_positions (List[Vector3D], optional): Initial positions of the points.
                Defaults to None.
            point_size (float, optional): _description_. Defaults to 1.0.
            rgb (Tuple[float, float, float], optional): RGB color code for the points.
                Defaults to (1, 0, 0).
            lifetime (float, optional): How long (in sec) the points should show. Defaults to 0.0
                (persistent).
            cid (int, optional): The client id to connect to the physics simulator. Defaults to 0.
        """

        self._lifetime = lifetime
        self._size = point_size
        self._rgb = rgb
        self._cid = cid

        self._points_id = -1
        if point_positions is not None:
            self._draw_points(point_positions=point_positions)

    def _draw_points(self, point_positions: List[Vector3D]):

        self._points_id = pb.addUserDebugPoints(
            pointPositions=point_positions,
            pointColorsRGB=[self._rgb] * len(point_positions),
            pointSize=self._size,
            lifeTime=self._lifetime,
            replaceItemUniqueId=self._points_id,
            physicsClientId=self._cid,
        )

    def update_point_positions(self, point_positions: List[Vector3D]):
        """Update the positions of the debug points in pybullet.

        Args:
            point_positions (List[Vector3D]): List of positions for the points in world.
        """
        self._draw_points(point_positions=point_positions)

    def remove(self):
        """Remove the debug points object from pybullet."""
        try:
            pb.removeUserDebugItem(self._points_id, physicsClientId=self._cid)
        except pb.error:
            pass



"""Defines state estimators that can be used in the control loop.
Currently, because we operate only in simulation, we directly use
data from the simulator and do not use any custom state estimation
algorithms."""

from .state_estimator_base import StateEstimatorBase, DummyStateEstimator



"""Dataclasses, enums and structs used in the control loop."""

from .tf_types import (
    Pose3D,
    Twist,
    RelativePose,
    Vector3D,
    QuatType,
)
from .robot_io import (
    EndEffectorStates,
    JointStates,
    StateEstimates,
    RobotState,
    RobotCmd,
)
from .motion_datatypes import (
    Vector3DTrajType,
    QuatTrajType,
    Trajectory3D,
    MultiFrameTrajectory3D,
    PointMotion,
    FrameMotion,
    GeneralisedTrajectory,
)
from .planner_outputs import LocalMotionPlan, GlobalMotionPlan, ControlMode, PlannerMode



from abc import ABC, abstractmethod
from typing import List

from ...utils.time_utils import RateTrigger, PythonPerfClock, ClockBase


class CustomCallbackBase(ABC):
    """Abstract base class for defining custom callbacks to be executed in the control loop."""

    @abstractmethod
    def run_once(self) -> None:
        """To be called in loop."""
        raise NotImplementedError("Method should be implemented in child class")

    def cleanup(self) -> None:
        """Override if custom cleaning up/shutting down is required."""
        return


class RateTriggeredMultiCallbacks(CustomCallbackBase):
    """Execute multiple `CustomCallbackBase` callbacks at the same specified (max) frequency."""

    def __init__(
        self,
        gui_callbacks: List[CustomCallbackBase],
        rate: float = None,
        clock: ClockBase = PythonPerfClock(),
    ):
        """Execute multiple `CustomCallbackBase` callbacks at the same specified (max) frequency.

        Args:
            gui_callbacks (List[CustomCallbackBase]): The CustomCallbackBase objects with their corresponding
                callbacks defined.
            rate (float, optional): The rate at which the callbacks for all the SliderCallback
                objects should be called. Defaults to None (call every time `run_once` is
                called).
            clock (ClockBase, optional): The ClockBase object to use for getting time. Defaults to
                PythonPerfClock().
        """
        self._callbacks: List[CustomCallbackBase] = []
        for sc in gui_callbacks:
            assert not isinstance(
                sc, RateTriggeredMultiCallbacks
            ), "Value cannot be of type `RateTriggeredMultiCallbacks`."
            self.add_callback(sc)
        if rate is not None:
            self._trigger = RateTrigger(rate=rate, clock=clock)
        else:
            self._trigger = lambda: True

    def add_callback(self, gui_callback: CustomCallbackBase):
        """Add an instance of PbGUISliderCallback or PbMultiGUISliderSingleCallback to the same rate
        trigger.

        Args:
            gui_callback (CustomCallbackBase): The GUI callback instance
                to add.
        """
        assert isinstance(
            gui_callback, CustomCallbackBase
        ), f"{gui_callback} is not of type `CustomCallbackBase."
        self._callbacks.append(gui_callback)

    def run_once(self):
        """This method has to be called for the slider callbacks to be executed. Will only
        execute if the trigger rate is met."""
        if self._trigger():
            for sc in self._callbacks:
                sc.run_once()

    def cleanup(self):
        for sc in self._callbacks:
            sc.cleanup()
        self._callbacks.clear()



from typing import List

from .controller_manager_base import ControllerManagerBase
from ...core.types import GlobalMotionPlan, RobotCmd, RobotState
from ..agents.agent_base import AgentBase
from .command_accumulators.cmd_accumulation_policies import (
    CommandAccumulatorBase,
    SimpleCmdOverride,
)
from ...core.logging import throttled_logging


class SimpleControllerManager(ControllerManagerBase):
    """A simple controller manager that naively performs sequential updates of all agents it
    manages."""

    def __init__(
        self,
        agents: List[AgentBase] = None,
        command_accumulation_policy: CommandAccumulatorBase = SimpleCmdOverride(),
    ):
        """Constructor.

        A simple controller manager that naively performs sequential updates of all agents it
        manages.

        Args:
            agents (List[AgentBase]): List of Agents to be managed by this manager.
            command_accumulation_policy (CommandAccumulatorBase, optional): The command accumulation
                policy to be used on the control outputs. Defaults to SimpleCmdOverride().
        """
        if agents is None:
            agents = []
        super().__init__(agents=agents, command_accumulation_policy=command_accumulation_policy)

    def update(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> RobotCmd:
        if len(self.agents) == 0:
            return RobotCmd()
        if len(self.agents) > 1 and isinstance(self._accumulator, SimpleCmdOverride):
            throttled_logging.warning(
                f"{self.__class__.__name__}: There are multiple agents in "
                "the control loop, but `SimpleCmdOverride` is used as the "
                "command accumulation policy. Only command from last agent "
                "will be used. Use `CmdMuxer` policy instead to comine "
                "commands from all agents.",
                1.0,
            )

        return self._accumulator.accumulate(
            commands=[
                agent.get_action(robot_state=robot_state, global_plan=global_plan, t=t, dt=dt)
                for agent in self.agents
            ]
        )


# NOTE: Ideal controller manager should handle controller switching, starting, stopping etc.



from abc import ABC, abstractmethod
import numpy as np

from ...core.types import GlobalMotionPlan, RobotCmd, RobotState, JointStates
from .agent_base import AgentBase


class MLAgentBase(AgentBase, ABC):
    """An abstract base class machine-learned controller agents.

    See docstrings for each method to be implemented.
    """

    @abstractmethod
    def initialise_robot_cmd(self, joint_states: JointStates):
        """Override this method in child class if custom initilisation is required (e.g. joint name order).
        By default, this method sets the joint name and position values to be equal to the input joint states
        object, with kp and kd set to be the default (in constructor).
        """

    @abstractmethod
    def update_input_to_model(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> np.ndarray:
        """Should update (`self._input_ndarray`) using appropriate values (input to model). This method
        has access to `self._latest_ctrl_cmd` (type `RobotCmd`) as well if needed.
        (NOTE: `self._latest_ctrl_cmd` is set to be the initial robot joint positions (zero velocities
        and efforts commands) with `default_kp` and `default_kd` at start).
        """

    @abstractmethod
    def update_cmd_from_model_output(
        self,
        model_output: np.ndarray,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float,
        dt: float,
    ) -> None:
        """Should update `self._latest_robot_cmd` (type `RobotCmd`) using the output from the NN model.

        Args:
            model_output (np.ndarray): the numpy array created from the output tensor from the model
                after the inference query was done. This is the output of the neural network. This
                method should use this object to update `self._latest_robot_cmd` to be sent to the
                robot.
        """

    @abstractmethod
    def get_action(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float = None,
        dt: float = None,
    ) -> RobotCmd:
        """Should return the control command given the current robot state and global plan.

        Args:
            robot_state (RobotState): The current state information from the robot.
            global_plan (GlobalMotionPlan): the latest global plan generated by the global planner
                used in the loop.
            t (float, optional): the current time signature of the control loop. Defaults to None
                (controllers may or may not need this).
            dt (float, optional): the time since the last control loop. Defaults to None
                (controllers may or may not need this).

        Returns:
            RobotCmd: The output control command to be sent to the robot.
        """



"""
Abstract base class for all core components in PyRCF control loop.

    Components that inherit this are:
        - RobotInterfaceBase
        - GlobalMotionPlannerBase
        - LocalMotionPlannerBase
        - ControllerBase
        - StateEstimatorBase
"""

from abc import ABC
from ..core.logging import logging


class PyRCFComponent(ABC):
    """Abstract base class for all core components in PyRCF control loop.

    Components that inherit this are:
        - RobotInterfaceBase
        - GlobalMotionPlannerBase
        - LocalMotionPlannerBase
        - ControllerBase
        - StateEstimatorBase
    """

    def shutdown(self):
        """Cleanly shutdown the PyRCF component. Override in child class if required. The
        base class implements an empty function."""
        logging.info(f"{self.__class__.__name__}: Shutting down.")

    def get_class_name(self) -> str:
        """Get the name of the class/type of this object."""
        return self.__class__.__name__

    def get_class_info(self) -> str:
        """Override with custom string if needed."""
        return self.get_class_name()



from dataclasses import dataclass, field
from enum import Enum

from .tf_types import Twist, RelativePose
from .robot_io import EndEffectorStates, JointStates
from .motion_datatypes import PointMotion


class ControlMode(Enum):
    """Request type from local planner to controller.

    MODES:
        - IDLE: controller should ignore any other values in the local plan.
        - CONTROL: controller can use values from the LocalMotionPlan to generate commands.
    """

    IDLE = 0
    """controller should ignore any other values in the local plan."""

    CONTROL = 1
    """controller will/should read rest of the LocalMotionPlan object only if in this mode."""


class PlannerMode(Enum):
    """Request type from global planner to local planner.

    MODES:
        - IDLE: No local plan should be generated by local planner.
        - CUSTOM: Local planner is free to use other attributes of the GlobalMotionPlan and
            generate a custom LocalMotionPlan message for the controller.
        - HOLD_POSITION: pre-defined action signal to keep the robot in the current position;
            other attributes in GlobalMotionPlan should be ignored by local planner, and planner
            should generate LocalMotionPlan that will keep the joints of the robot locked.
    """

    IDLE = 0
    """No local plan should be generated. No command is to generated by controller in this state."""

    HOLD_POSITION = 1
    """pre-defined action signal to keep the robot in the current position, other attributes in
    GlobalMotionPlan should be ignored by local planner (this state has to be handled by local
    planner or control loop; see examples.03).
    """

    CUSTOM = 2
    """In this mode, the local planner is free to use other attributes of the GlobalMotionPlan and
    generate a custom LocalMotionPlan message for the controller.
    """


@dataclass
class GlobalMotionPlan:
    """Object representing a global motion plan provided by a generic global motion planner.

    In our case, parsed user input is translated to this format for local planner.

    In full autonomous robot framework, this should be generated by a global world-aware planner,
    based on task.


    Attributes (see doc for each attribute):
        relative_pose (RelativePose): Relative pose target in teleop frame.
        twist (Twist): target velocity command for the base in base frame.
        joint_references (JointStates): joint targets.
        end_effector_references (EndEffectorStates): references for end-effectors of the robot.
        planner_mode (PlannerMode): available modes: [PlannerMode.IDLE, PlannerMode.HOLD_POSITION,
            PlannerMode.CUSTOM]
    """

    relative_pose: RelativePose = None
    """Pose in teleop frame of the robot.

    Teleop frame is equivalent to robot (base) frame, but has the roll, pitch and height values
    fixed in some neutral pre-defined values in the world coordinates.

    We define the following as our teleop frame. This may be changed as desired.
    [x,y,z] = [x_BF, y_BF, 0.0] in world
    [r,p,y] = [0.0, 0.0, yaw_BF] in world
    """

    twist: Twist = field(default_factory=Twist)
    """Twist in base/teleop frame of the robot"""

    # joint reference is not ideal in a global plan message; only left for directly sending
    # reference using joystick/gui (for testing new robots, or for controlling manipulator)
    joint_references: JointStates = field(default_factory=JointStates)
    """Desired joint references. Typically a local planner would only fill joint position or
    velocity values, and not effort values."""
    # ee references is not ideal in a global plan message; only left for directly sending reference
    # using joystick/gui (for testing new robots, or for controlling manipulator)
    end_effector_references: EndEffectorStates = field(default_factory=EndEffectorStates)
    """Desired end-effector pose (in global frame), contact states, contact forces etc. can be
    defined here."""

    planner_mode: PlannerMode = PlannerMode.IDLE
    """Defined by a global planner to let the local planner know if the rest of the
    GlobalMotionPlan message is usable, or if there is a specific plan to be generated (such as
    HOLD_POSITION)"""


@dataclass
class LocalMotionPlan:
    """Object representing a local motion plan provided by a generic local motion planner.

    This object will be used as reference by controllers.

    Attributes (see doc for each attribute):
        relative_pose (RelativePose): Relative pose target in teleop frame.
        twist (Twist): target velocity command for the base in base frame.
        joint_references (JointStates): references for joints (position, velocity)
        end_effector_references (EndEffectorStates): References for end-effectors (pos, force,
            contact states)
        control_mode (ControlMode): control modes (available modes: [ControlMode.IDLE,
            ControlMode.CONTROL])

    WARNING
    -------
    NOTE: as of now, the base pose (`relative_pose`) is defined in the teleop frame, and
    the end-effector poses (in `end_effector_references.ee_poses`) is in the world frame.
    This may have to be changed later if there is a need.
    """

    # setting default value of identity is wrong (due to definition of teleop frame)
    relative_pose: RelativePose = None
    """Pose in teleop frame of the robot.

    Teleop frame is equivalent to robot (base) frame, but has the roll, pitch and height values
    fixed in some neutral pre-defined values in the world coordinates.

    We define the following as our teleop frame. This may be changed as desired.
    [x,y,z] = [x_BF, y_BF, 0.0] in world
    [r,p,y] = [0.0, 0.0, yaw_BF] in world

    WARNING
    -------
    NOTE: The frame used may have to be changed later. We may need to use this in the global frame.
    """
    twist: Twist = field(default_factory=Twist)
    """Base velocity commands in the base frame."""
    joint_references: JointStates = field(default_factory=JointStates)
    """Desired joint references. Typically a local planner would only fill joint position or
    velocity values, and not effort values."""
    end_effector_references: EndEffectorStates = field(default_factory=EndEffectorStates)
    """Desired end-effector pose (in global frame), contact states, contact forces etc. can be
    defined here.

    WARNING
    -------
    NOTE: The frame used may have to be changed later, depending on whether it is more useful in
    global/base/teleop frame.
    """
    com_reference: PointMotion = None
    """Reference position ([x,y,z] in global frame), velocity and acceleration for center of
    mass of the robot for whole-body controllers."""
    control_mode: ControlMode = ControlMode.IDLE
    """Defined by a local planner to let the controller(s) know if the rest of the LocalMotionPlan
    message is usable, or if there is a specific control action to take (such as IDLE)."""



"""Helper functions for updating provided GlobalMotionPlan objects for UI Interfaces."""

from typing import Callable, Dict, List, Any
from functools import partial
import numpy as np

from ....core.types import GlobalMotionPlan, PlannerMode
from ....core.logging import logging

# pylint: disable=C0116,C0103


class _GlobalPlanModifiers:
    """Helper functions for updating provided GlobalMotionPlan objects (internal use only)."""

    @staticmethod
    def set_linear_velocity(
        global_plan: GlobalMotionPlan,
        linear_velocity: np.ndarray,
        scaling: np.ndarray,
        prev_vel_scaling: np.ndarray,
    ) -> GlobalMotionPlan:
        assert len(linear_velocity) == 3
        global_plan.twist.linear = (
            prev_vel_scaling * global_plan.twist.linear + scaling * linear_velocity
        )
        return global_plan

    @staticmethod
    def set_rotational_velocity(
        global_plan: GlobalMotionPlan,
        rotational_velocity: np.ndarray,
        scaling: np.ndarray,
        prev_vel_scaling: np.ndarray,
    ) -> GlobalMotionPlan:
        assert len(rotational_velocity) == 3
        global_plan.twist.angular = (
            prev_vel_scaling * global_plan.twist.angular + scaling * rotational_velocity
        )
        return global_plan

    @staticmethod
    def set_plan_mode(global_plan: GlobalMotionPlan, planner_mode: PlannerMode):
        global_plan.planner_mode = planner_mode
        return global_plan

    @staticmethod
    def toggle_plan_mode(global_plan: GlobalMotionPlan, planner_modes: List[PlannerMode]):
        if global_plan.planner_mode != planner_modes[0]:
            global_plan.planner_mode = planner_modes[0]
        else:
            global_plan.planner_mode = planner_modes[1]
        return global_plan

    @staticmethod
    def increment_linear_velocity(
        global_plan: GlobalMotionPlan, linear_velocity: np.ndarray
    ) -> GlobalMotionPlan:
        assert len(linear_velocity) == 3
        global_plan.twist.linear += np.array(linear_velocity)
        return global_plan

    @staticmethod
    def increment_rotational_velocity(
        global_plan: GlobalMotionPlan, rotational_velocity: np.ndarray
    ) -> GlobalMotionPlan:
        assert len(rotational_velocity) == 3
        global_plan.twist.angular += np.array(rotational_velocity)
        return global_plan

    @staticmethod
    def increment_relative_position(global_plan: GlobalMotionPlan, position_delta: np.ndarray):
        assert len(position_delta) == 3
        global_plan.relative_pose.position += np.array(position_delta)
        return global_plan

    @staticmethod
    def increment_joint_position(
        global_plan: GlobalMotionPlan, joint_id: int, position_delta: float
    ):
        if global_plan.joint_references.joint_positions is not None:
            global_plan.joint_references.joint_positions[joint_id] += position_delta
        return global_plan


class GlobalPlanJointPositionIncrementer:
    """Helper to bind a button to change the value of different joints incrementally.
    One button can be used to increment the value of the joint by a specific number.
    Another button(s) can be used to change the index of the joint being modified.
    """

    def __init__(self, update_only_if_planner_activated: bool = True):
        self._check_and_update = update_only_if_planner_activated
        self.joint_id = 0

    def _check_ready(self, global_plan: GlobalMotionPlan):
        if self._check_and_update and global_plan.planner_mode != PlannerMode.CUSTOM:
            logging.warning(
                f"{self.__class__.__name__}: Planner is not activated. Not updating joint commands."
                " Set PlannerMode to PlannerMode.CUSTOM in the global plan."
            )
            return False
        return True

    def increment_joint_target(self, global_plan: GlobalMotionPlan, position_delta: float):
        """Use as binding for a button in the format:
        `{ <button_id>: partial(obj.increment_joint_target, position_delta=value) }`"""
        if not self._check_ready(global_plan=global_plan):
            return global_plan
        if global_plan.joint_references.joint_positions is not None:
            global_plan.joint_references.joint_positions[self.joint_id] += position_delta
        return global_plan

    def increase_joint_id(self, global_plan: GlobalMotionPlan):
        """
        Change the id of the joint being modified by incrementing the index by 1.
        Use the handle to this function directly as binding for a button in the format:
        `{ <button_id>: obj.increase_joint_id }`"""
        if not self._check_ready(global_plan=global_plan):
            return global_plan
        if global_plan.joint_references.joint_positions is None:
            return global_plan
        if self.joint_id >= len(global_plan.joint_references.joint_positions) - 1:
            self.joint_id = 0
        else:
            self.joint_id += 1
        return global_plan

    def decrease_joint_id(self, global_plan: GlobalMotionPlan):
        """
        Change the id of the joint being modified by reducing the index by 1.
        Use the handle to this function directly as binding for a button in the format:
        `{ <button_id>: obj.decrease_joint_id }`"""
        if not self._check_ready(global_plan=global_plan):
            return global_plan
        if global_plan.joint_references.joint_positions is None:
            return global_plan
        if self.joint_id <= 0:
            self.joint_id = len(global_plan.joint_references.joint_positions) - 1
        else:
            self.joint_id -= 1
        return global_plan


class Hold2SwapBindings:
    """Helper tool to make holding a button change the behaviour of another button/trigger."""

    def __init__(
        self,
        default_binding: Callable[[GlobalMotionPlan, Any], GlobalMotionPlan],
        alternate_binding: Callable[[GlobalMotionPlan, Any], GlobalMotionPlan],
    ):
        """Helper tool to make holding a button (called "B" here) change the behaviour of another
        button/trigger (called "button A" here).

        Args:
            default_binding (Callable[[GlobalMotionPlan, Any], GlobalMotionPlan]): The default
                behaviour of the binded button ("button A").
            alternate_binding (Callable[[GlobalMotionPlan, Any], GlobalMotionPlan]): Alternate
                behaviour of the binded button ("button A") when button "button B" is held.

        """
        self._default_binding = default_binding
        self._alternate_binding = alternate_binding
        self._trigger_held = False

    def trigger_swap(self, _any: Any):
        """Set this function as the binding for the trigger hold button ("button B")."""
        self._trigger_held = True
        return _any

    def use_binding(self, *args, **kwargs):
        """Set this function as the binding for the main button ("button A")."""
        if self._trigger_held:
            out = self._alternate_binding(*args, **kwargs)
        else:
            out = self._default_binding(*args, **kwargs)
        self._trigger_held = False
        return out


class Key2GlobalPlanHelper:
    """Helper class for creating keyboard mapppings for keyboard_interface.

    Provides static methods/objects that can be used for easily creating lambdas for creating
    key mappings (to be used for keyboard_interface)

    Each of the following will return a callable in the appropriate signature and can be used
    directly for the keyboard_interface. (see DEFAULT_KEYBOARD_MAPPING defined below)

    Available objects/methods (static):
        - CUSTOM
        - IDLE
        - HOLD_POSITION
        - APPLY_LIN_VEL(linear_velocity: np.ndarray)
        - APPLY_ROT_VEL(rotational_velocity: np.ndarray)
        - APPLY_LIN_VEL_DELTA(linear_velocity: np.ndarray)
        - APPLY_ROT_VEL_DELTA(rotational_velocity: np.ndarray)
    """

    CUSTOM: Callable[[GlobalMotionPlan], GlobalMotionPlan] = partial(
        _GlobalPlanModifiers.set_plan_mode, planner_mode=PlannerMode.CUSTOM
    )
    """static object that returns a callable. Sets the planner_mode to CUSTOM (controller
    activate)."""

    IDLE: Callable[[GlobalMotionPlan], GlobalMotionPlan] = partial(
        _GlobalPlanModifiers.set_plan_mode, planner_mode=PlannerMode.IDLE
    )
    """static object that returns a callable. Sets the planner_mode to IDLE."""

    HOLD_POSITION: Callable[[GlobalMotionPlan], GlobalMotionPlan] = partial(
        _GlobalPlanModifiers.set_plan_mode, planner_mode=PlannerMode.HOLD_POSITION
    )
    """static object that returns a callable. Sets the planner_mode to HOLD_POSITION."""

    TOGGLE_CUSTOM_IDLE: Callable[[GlobalMotionPlan], GlobalMotionPlan] = partial(
        _GlobalPlanModifiers.toggle_plan_mode,
        planner_modes=[PlannerMode.CUSTOM, PlannerMode.IDLE],
    )

    @staticmethod
    def APPLY_LIN_VEL(
        linear_velocity: np.ndarray,
    ) -> Callable[[GlobalMotionPlan], GlobalMotionPlan]:
        """Static method that returns a callable for setting linear velocity (takes one argument).

        Args:
            linear_velocity (np.ndarray): linear velocity (in base frame) to be set (m/sec)

        Returns:
            Callable[[GlobalMotionPlan], GlobalMotionPlan]: returns a callable that can be used by
                the keyboard_interface.
        """
        return partial(
            _GlobalPlanModifiers.set_linear_velocity,
            linear_velocity=linear_velocity,
            scaling=1,
            prev_vel_scaling=0,
        )

    @staticmethod
    def APPLY_ROT_VEL(
        rotational_velocity: np.ndarray,
    ) -> Callable[[GlobalMotionPlan], GlobalMotionPlan]:
        """Static method that returns a callable for setting rotational velocity (takes one
        argument).

        Args:
            rotational_velocity (np.ndarray): rotational velocity (in base frame) to be set
                (rad/sec)

        Returns:
            Callable[[GlobalMotionPlan], GlobalMotionPlan]: returns a callable that can be used by
                the keyboard_interface.
        """
        return partial(
            _GlobalPlanModifiers.set_rotational_velocity,
            rotational_velocity=rotational_velocity,
            scaling=1,
            prev_vel_scaling=0,
        )

    @staticmethod
    def APPLY_LIN_VEL_DELTA(
        linear_velocity: np.ndarray,
    ) -> Callable[[GlobalMotionPlan], GlobalMotionPlan]:
        """Static method that returns a callable for INCREMENTING current linear velocity (takes
            one argument).

        Args:
            linear_velocity (np.ndarray): linear velocity (in base frame) to be added (can be
                negative) (m/sec)

        Returns:
            Callable[[GlobalMotionPlan], GlobalMotionPlan]: returns a callable that can be used by
                the keyboard_interface.
        """
        return partial(
            _GlobalPlanModifiers.increment_linear_velocity,
            linear_velocity=linear_velocity,
        )

    @staticmethod
    def APPLY_ROT_VEL_DELTA(
        rotational_velocity: np.ndarray,
    ) -> Callable[[GlobalMotionPlan], GlobalMotionPlan]:
        """Static method that returns a callable for INCREMENTING current rotational velocity
            (takes one argument).

        Args:
            rotational_velocity (np.ndarray): rotational velocity (in base frame) to be added
                (can be negative) (rad/sec)

        Returns:
            Callable[[GlobalMotionPlan], GlobalMotionPlan]: returns a callable that can be used by
                the keyboard_interface.
        """
        return partial(
            _GlobalPlanModifiers.increment_rotational_velocity,
            rotational_velocity=rotational_velocity,
        )

    @staticmethod
    def APPLY_RELATIVE_POS_DELTA(
        position_delta: np.ndarray,
    ) -> Callable[[GlobalMotionPlan], GlobalMotionPlan]:
        """Static method that returns a callable for INCREMENTING position in teleop frame (takes
        one argument).

        Args:
            position_delta (np.ndarray): values to be added to current position of the robot base
                in the teleop frame

        Returns:
            Callable[[GlobalMotionPlan], GlobalMotionPlan]: returns a callable that can be used by
                the keyboard_interface.
        """
        return partial(
            _GlobalPlanModifiers.increment_relative_position,
            position_delta=position_delta,
        )

    @staticmethod
    def APPLY_JOINT_POS_DELTA(
        joint_id: int,
        position_delta: float,
    ) -> Callable[[GlobalMotionPlan], GlobalMotionPlan]:
        """Static method that returns a callable for INCREMENTING joint position for the specified
        joint id.

        Args:
            position_delta (float): value to be added to current joint position

        Returns:
            Callable[[GlobalMotionPlan], GlobalMotionPlan]: returns a callable that can be used by
                the keyboard_interface.
        """
        return partial(
            _GlobalPlanModifiers.increment_joint_position,
            joint_id=joint_id,
            position_delta=position_delta,
        )


class Analog2GlobalPlanHelper:
    """Helper class for creating mapppings for non-binary analog keys (e.g. analog gamepad stick).

    Provides static methods/objects that can be used for easily creating lambdas for creating
    key mappings (to be used for non-binary analog keys)

    Each of the following will return a callable in the appropriate signature and can be used
    directly for the non-binary analog keys. (see DEFAULT_KEYBOARD_MAPPING defined below)

    Available objects/methods (static):
        - CUSTOM
        - IDLE
        - HOLD_POSITION
        - APPLY_LIN_VEL(linear_velocity: np.ndarray)
        - APPLY_ROT_VEL(rotational_velocity: np.ndarray)
        - APPLY_LIN_VEL_DELTA(linear_velocity: np.ndarray)
        - APPLY_ROT_VEL_DELTA(rotational_velocity: np.ndarray)
    """

    @staticmethod
    def APPLY_LIN_VEL_SCALED(
        max_linear_velocity: np.ndarray,
        prev_vel_scaling: np.ndarray,
    ) -> Callable[[GlobalMotionPlan], GlobalMotionPlan]:
        """Static method that returns a callable for setting linear velocity (takes 2 arguments).

        Args:
            max_linear_velocity (np.ndarray): max linear velocity (in base frame) to be set (m/sec)
            prev_vel_scaling (np.ndarray): scaling to use on previous velocity. This is added to
                the new scaled linear velocity input.

        Returns:
            Callable[[GlobalMotionPlan, np.ndarray, np.ndarray], GlobalMotionPlan]: returns a
                callable that can be used by the joystick_interface with 2 arguments (global
                plan, analog np.ndarray scale value for each axis, np.ndarray scale value for
                previous value in the global plan message).
        """
        return lambda global_plan, analog_axis_scale: _GlobalPlanModifiers.set_linear_velocity(
            global_plan=global_plan,
            linear_velocity=max_linear_velocity,
            scaling=analog_axis_scale,
            prev_vel_scaling=prev_vel_scaling,
        )

    @staticmethod
    def APPLY_ROT_VEL_SCALED(
        max_rotational_velocity: np.ndarray,
        prev_vel_scaling: np.ndarray,
    ) -> Callable[[GlobalMotionPlan], GlobalMotionPlan]:
        """Static method that returns a callable for setting rotational velocity (takes 2
        arguments).

        Args:
            max_rotational_velocity (np.ndarray): max rotational velocity (in base frame) to be set
                (m/sec)
            prev_vel_scaling (np.ndarray): scaling to use on previous velocity. This is added to
                the new scaled rotational velocity input.

        Returns:
            Callable[[GlobalMotionPlan, np.ndarray, np.ndarray], GlobalMotionPlan]: returns a
                callable that can be used by the joystick_interface with 2 arguments (global plan,
                analog np.ndarray scale value for each axis, np.ndarray scale value for previous
                value in the global plan message).
        """
        return lambda global_plan, analog_axis_scale: _GlobalPlanModifiers.set_rotational_velocity(
            global_plan=global_plan,
            rotational_velocity=max_rotational_velocity,
            scaling=analog_axis_scale,
            prev_vel_scaling=prev_vel_scaling,
        )


def get_keymapping_doc(
    key_mappings: Dict[str, Callable[[GlobalMotionPlan], GlobalMotionPlan]],
):
    """Get a docstring for a given Mapping from string to callable."""
    out_string = "Key mappings:\n"
    docs = {}
    for k in key_mappings:
        if isinstance(key_mappings[k], partial):
            doc = f"{key_mappings[k].func.__name__}:\n\t\targs: {key_mappings[k].keywords}"
        else:
            doc = key_mappings[k].__name__
        docs[k] = doc
        out_string += f"\t{k}: {doc}\n"

    return out_string, docs



import copy
from typing import List, Mapping
import numpy as np
import pybullet_robot
from pybullet_robot.utils.robot_loader_utils import get_urdf_from_awesome_robot_descriptions

from .sim_robot_base import SimulatedRobotInterface
from ....core.types import (
    Pose3D,
    RobotState,
    StateEstimates,
    RobotCmd,
    Twist,
    EndEffectorStates,
    QuatType,
    JointStates,
)
from ....utils.math_utils import quat2rot
from ....utils.time_utils import ClockBase
from ....core.logging import logging
from ....utils.urdf_utils import (
    temp_path_urdf_with_joints_fixed,
    create_new_urdf_with_joints_fixed,
)

# pylint: disable = C0103


class PybulletRobot(SimulatedRobotInterface):
    """A RobotInterface implementation for a generic robot in pybullet.

    All subclasses of this method will use synchronous sim time stepping, i.e.
    the simulated world will step only when the read() method of this class is
    called. This is to keep the clock synchronous and for providing full repeatable
    performance to the control loop by minimising parallelisation.
    """

    DEFAULT_URDF_PATH: str = None
    """Default urdf path to be used for particular instances of PybulletRobot. This
    variable should be overridden in the implemented child classes."""

    class BulletWorldClock(ClockBase):
        """Only valid if sim stepping is done synchronously and managed by the PybulletRobot class."""

        def __init__(self, dt: float) -> None:
            self._current_t = 0.0
            self._dt = dt

        def step_time(self):
            """To be called each time simulation steps forward."""
            self._current_t += self._dt

        def get_time(self) -> float:
            return self._current_t

    def __init__(
        self,
        urdf_path: str,
        ee_names: List[str] = None,
        place_on_ground: bool = True,
        default_base_position: np.ndarray = np.zeros(3),
        default_base_orientation: np.ndarray = np.array([0, 0, 0, 1]),
        default_joint_positions: np.ndarray = None,
        create_pinocchio_interface: bool = True,
        floating_base: bool = True,
        verbose: bool = True,
        **sim_interface_kwargs,
    ):
        """A RobotInterface implementation for a generic robot in pybullet.

        All subclasses of this method will use synchronous sim time stepping, i.e.
        the simulated world will step only when the read() method of this class is
        called. This is to keep the clock synchronous and for providing full repeatable
        performance to the control loop by minimising parallelisation.

        Args:
            urdf_path (str): Path to urdf file of robot.
            ee_names (List[str], optional): List of end-effectors for the robot. Defaults to None.
            place_on_ground (bool): If true, the base position height will automatically be
                adjusted such that the robot is on the ground with the given joint positions.
                Defaults to True.
            default_base_position (np.ndarray, optional): Default position of the base of the robot
                in the world frame during start-up. Note that the height value (z) is only used if
                'place_on_ground' is set to False. Defaults to np.zeros(3).
            default_base_orientation (np.ndarray, optional): Default orientation quaternion in the
                world frame for the base of the robot during start-up. Defaults to
                np.array([0, 0, 0, 1]).
            default_joint_positions (List[float], optional): Optional starting values for joints.
                Defaults to None. NOTE: These values should be in the order of joints in pybullet.
            create_pinocchio_interface (bool, optional): If true, creates a PinocchioInterface
                object for this robot. The object is automatically updated with latest state when
                read() is called. Defaults to True.
            floating_base (bool, optional): Specifying whether this robot has a floating or fixed
                base (setting to True creates a floating base instance with pinocchio, and makes
                the simulated robot free floating (not fixed base)). Defaults to True.
            verbose (bool, optional): Verbosity flag for debugging robot info during construction.
                Defaults to True.

            **sim_interface_kwargs: additional keyword arguments to pass down to the internal
                Simulator interface (i.e. additional arguments to pybullet_robot.BulletRobot).
        """
        if urdf_path is None:
            urdf_path = self.DEFAULT_URDF_PATH

        sim_interface_kwargs["enable_torque_mode"] = sim_interface_kwargs.get(
            "enable_torque_mode", True
        )

        if ee_names is None:
            ee_names = []

        self._sim_robot = pybullet_robot.BulletRobot(
            urdf_path=urdf_path,
            run_async=False,  # this interface will only allows synchronous sim stepping
            ee_names=ee_names,
            default_joint_positions=default_joint_positions,
            place_on_ground=place_on_ground,
            default_base_position=default_base_position,
            default_base_orientation=default_base_orientation,
            verbose=verbose,
            use_fixed_base=(not floating_base),
            **sim_interface_kwargs,
        )

        # this is only usable because we use synchronous time stepping in sim
        self._clock = self.BulletWorldClock(
            dt=self._sim_robot.get_physics_parameters()["fixedTimeStep"]
        )

        super().__init__(
            robot_urdf=urdf_path,
            ee_names=ee_names,
            floating_base=floating_base,
            verbose=verbose,
            create_pinocchio_interface=create_pinocchio_interface,
        )
        if self.has_pinocchio_interface:
            sim_robot_state = copy.deepcopy(self._sim_robot.get_robot_states())
            rot_mat = quat2rot(quaternion=sim_robot_state["base_quaternion"]).T
            state = RobotState(
                joint_states=JointStates(
                    joint_names=self._sim_robot.actuated_joint_names,
                    joint_positions=sim_robot_state["actuated_joint_positions"],
                    joint_velocities=sim_robot_state["actuated_joint_velocities"],
                    joint_efforts=sim_robot_state["actuated_joint_torques"],
                ),
                state_estimates=StateEstimates(
                    pose=Pose3D(
                        position=sim_robot_state["base_position"],
                        orientation=sim_robot_state["base_quaternion"],
                    ),
                    twist=Twist(  # twist represented in base frame
                        linear=rot_mat @ sim_robot_state["base_velocity_linear"],
                        angular=rot_mat @ sim_robot_state["base_velocity_angular"],
                    ),
                    end_effector_states=EndEffectorStates(
                        contact_states=sim_robot_state["ee_contact_states"],
                        ee_names=sim_robot_state["ee_order"],
                    ),
                ),
            )
            self.update_pinocchio_robot_state(robot_state=state)

    def read(self) -> RobotState:
        self._sim_robot.step()  # synchronised sim stepping
        self._clock.step_time()

        sim_robot_state = copy.deepcopy(self._sim_robot.get_robot_states())

        try:
            rot_mat = quat2rot(quaternion=sim_robot_state["base_quaternion"]).T
        except ValueError:
            print(f"Sim quaternion: {sim_robot_state['base_quaternion']}")
            raise

        state = RobotState(
            joint_states=JointStates(
                joint_names=self._sim_robot.actuated_joint_names,
                joint_positions=sim_robot_state["actuated_joint_positions"],
                joint_velocities=sim_robot_state["actuated_joint_velocities"],
                joint_efforts=sim_robot_state["actuated_joint_torques"],
            ),
            state_estimates=StateEstimates(
                pose=Pose3D(
                    position=sim_robot_state["base_position"],
                    orientation=sim_robot_state["base_quaternion"],
                ),
                twist=Twist(  # twist represented in base frame
                    linear=rot_mat @ sim_robot_state["base_velocity_linear"],
                    angular=rot_mat @ sim_robot_state["base_velocity_angular"],
                ),
                end_effector_states=EndEffectorStates(
                    contact_states=sim_robot_state["ee_contact_states"],
                    ee_names=sim_robot_state["ee_order"],
                ),
            ),
        )

        if self.has_pinocchio_interface:
            state = self.update_pinocchio_robot_state(robot_state=state, update_ee_states=True)

        return state

    def write(self, cmd: RobotCmd) -> bool:
        if (jnames := cmd.joint_commands.joint_names) is None:
            return True
        try:
            self._sim_robot.set_actuated_joint_commands(
                actuated_joint_names=(jnames),
                q=(
                    cmd.joint_commands.joint_positions
                    if cmd.joint_commands.joint_positions is not None
                    else np.zeros(len(jnames))
                ),
                Kp=cmd.Kp if cmd.Kp is not None else np.zeros(len(jnames)),
                dq=(
                    cmd.joint_commands.joint_velocities
                    if cmd.joint_commands.joint_velocities is not None
                    else np.zeros(len(jnames))
                ),
                Kd=cmd.Kd if cmd.Kd is not None else np.zeros(len(jnames)),
                tau=(
                    cmd.joint_commands.joint_efforts
                    if cmd.joint_commands.joint_efforts is not None
                    else np.zeros(len(jnames))
                ),
            )
            return True
        except Exception as e:  # pylint:disable=W0718
            logging.exception(
                f"{__class__.__name__}: Control command could not be written to simulator. {e}"
                f" Culprit {cmd}. len joints: {len(self._sim_robot.actuated_joint_names)}."
            )
            return False

    def set_base_pose(self, position: np.ndarray, orientation: QuatType):
        self._sim_robot.reset_base_pose(position=position, orientation=orientation)

    def get_sim_clock(self) -> ClockBase:
        return self._clock

    def shutdown(self):
        super().shutdown()
        if hasattr(self, "_sim_robot"):
            self._sim_robot.shutdown()

    @classmethod
    def withJointsLocked(
        cls: "PybulletRobot",
        joints_to_lock: Mapping[str, float] = None,
        temp_urdf: bool = True,
        **kwargs,
    ):
        """Classmethod for RobotInterface implementation for PybulletRobot robot in pybullet, that
        locks the specified joints of the robot. All specified joints are converted to "fixed"
        joint type and cannot be reverted dynamically!

        NOTE: Use this method with caution. All keyword arguments are directly passed to the
        default constructor!

        Args:
            joints_to_lock (Mapping[str, float]): Dictionary containing name of joints that should
                be locked as keys, and the joint values to lock them at as values. By default, uses
                upper body joints (from upper torso upwards) set to zeros.
            temp_urdf (bool, optional): If set to True, will create a temporary urdf that will be
                deleted after loading in pybullet, otherwise create a file in /tmp/ directory.
                Defaults to True.

            **kwargs: additional keyword arguments to pass to PybulletRobot or subclass.


        -------------------------------------------------------------------------------------------
        NOTE
        ----
        By default, the urdf created via this constructor is only valid during construction. The
        path to the urdf (accessed through self._sim_robot.urdf_path, for instance) will not be a
        valid path once the init is completed. This is because this method temporarily creates a
        urdf after modifying the specified joints to "fixed" joint type, and destroys the temporary
        file soon after all objects using it (pybullet, pinocchio, etc.) are created in the
        construction.

        To disable this behaviour, set `temp_urdf=False` during construction. This will create a
        permenant urdf file in the /tmp/ directory.
        -------------------------------------------------------------------------------------------
        """
        if joints_to_lock is None:
            joints_to_lock = {}

        urdf_path = kwargs.get("urdf_path", cls.DEFAULT_URDF_PATH)
        kwargs.pop("urdf_path", None)
        if temp_urdf:
            with temp_path_urdf_with_joints_fixed(urdf_path, joints_to_lock) as updated_urdf:
                return cls(urdf_path=updated_urdf, **kwargs)

        updated_urdf = create_new_urdf_with_joints_fixed(urdf_path, joints_to_lock)
        return cls(urdf_path=updated_urdf, **kwargs)

    @classmethod
    def fromAwesomeRobotDescriptions(
        cls: "PybulletRobot",
        robot_description_name: str,
        ee_names: List[str] = None,
        place_on_ground: bool = True,
        default_base_position: np.ndarray = np.zeros(3),
        default_base_orientation: np.ndarray = np.array([0, 0, 0, 1]),
        default_joint_positions: np.ndarray = None,
        create_pinocchio_interface: bool = True,
        floating_base: bool = True,
        verbose: bool = True,
        **sim_interface_kwargs,
    ) -> "PybulletRobot":
        """Create a PybulletRobot instance using robots available in the open source Awesome Robot
        Descriptions list
        (https://github.com/robot-descriptions/robot_descriptions.py/tree/main?tab=readme-ov-file#descriptions).


        The list of available robot descriptions can also be viewed in the variable `AWESOME_ROBOTS`
        imported from `pybullet_robot.utils.urdf_utils`.

        Downloads description package for the specified robot and caches it locally (only needs
        downloading once), and loads this robot as a PybulletRobot instance. NOTE: `ee_names` should
        be provided if proper end-effector kinematics is required. Also starting pose and joint
        configurations are always at zero. So to set to some specific configuration provide
        appropriate arguments.

        Args:
            robot_description_name (str): The name of the robot description package name. Should be
                a value available in the awesome robots list (also defined in variable
                `AVAILABLE_ROBOTS` in `utils.urdf_utils`).
            ee_names (List[str], optional): List of end-effectors for the robot. Defaults to None.
            place_on_ground (bool): If true, the base position height will automatically be adjusted
                such that the robot is on the ground with the given joint positions. Defaults to
                True.
            default_base_position (np.ndarray, optional): Default position of the base of the robot
                in the world frame during start-up. Note that the height value (z) is only used if
                'place_on_ground' is set to False. Defaults to np.zeros(3).
            default_base_orientation (np.ndarray, optional): Default orientation quaternion in the
                world frame for the base of the robot during start-up. Defaults to
                np.array([0, 0, 0, 1]).
            default_joint_positions (List[float], optional): Optional starting values for joints.
                Defaults to None. NOTE: These values should be in the order of joints in pybullet.
            create_pinocchio_interface (bool, optional): If true, creates a PinocchioInterface
                object for this robot. The object is automatically updated with latest state when
                read() is called. Defaults to True.
            floating_base (bool, optional): Specifying whether this robot has a floating or fixed
                base (setting to True creates a floating base instance with pinocchio, and makes the
                simulated robot free floating (not fixed base)). Defaults to True.
            verbose (bool, optional): Verbosity flag for debugging robot info during construction.
                Defaults to True.

        Returns:
            PybulletRobot: A PybulletRobot instance created using the URDF generated using the awesome
                robot descriptions package.
        """
        return cls(
            urdf_path=get_urdf_from_awesome_robot_descriptions(
                robot_description_pkg_name=robot_description_name
            ),
            ee_names=ee_names,
            place_on_ground=place_on_ground,
            default_base_position=default_base_position,
            default_base_orientation=default_base_orientation,
            default_joint_positions=default_joint_positions,
            create_pinocchio_interface=create_pinocchio_interface,
            floating_base=floating_base,
            verbose=verbose,
            **sim_interface_kwargs,
        )



from typing import Callable, Any, Tuple
from functools import partial
from dataclasses import dataclass, field
import numpy as np

from ....core.types import GlobalMotionPlan
from .ui_utils import (
    Key2GlobalPlanHelper,
    Analog2GlobalPlanHelper,
    GlobalPlanJointPositionIncrementer,
)

MAX_LIN_VEL = 0.5
MAX_ANG_VEL = 1.0
LIN_VEL_DELTA = 0.05
ANG_VEL_DELTA = 0.1
POS_DELTA = 0.02
JOINT_POS_DELTA = 0.01


_jpi = GlobalPlanJointPositionIncrementer()

DEFAULT_KEYBOARD_MAPPING = {
    "o": Key2GlobalPlanHelper.TOGGLE_CUSTOM_IDLE,
    "h": Key2GlobalPlanHelper.HOLD_POSITION,
    "w": Key2GlobalPlanHelper.APPLY_LIN_VEL_DELTA(np.array([LIN_VEL_DELTA, 0.0, 0.0])),
    "s": Key2GlobalPlanHelper.APPLY_LIN_VEL_DELTA(np.array([-LIN_VEL_DELTA, 0.0, 0.0])),
    "q": Key2GlobalPlanHelper.APPLY_LIN_VEL_DELTA(np.array([0.0, LIN_VEL_DELTA, 0.0])),
    "e": Key2GlobalPlanHelper.APPLY_LIN_VEL_DELTA(np.array([0.0, -LIN_VEL_DELTA, 0.0])),
    "a": Key2GlobalPlanHelper.APPLY_ROT_VEL_DELTA(np.array([0.0, 0.0, ANG_VEL_DELTA])),
    "d": Key2GlobalPlanHelper.APPLY_ROT_VEL_DELTA(np.array([0.0, 0.0, -ANG_VEL_DELTA])),
    "up": partial(_jpi.increment_joint_target, position_delta=JOINT_POS_DELTA),
    "down": partial(_jpi.increment_joint_target, position_delta=-JOINT_POS_DELTA),
    "right": _jpi.increase_joint_id,
    "left": _jpi.decrease_joint_id,
}
"""Default key mappings when using the keyboard interface for sending global motion
plan messages.

- o --> TOGGLE between CUSTOM and IDLE
- h --> HOLD_POSITION
- w --> APPLY_LIN_VEL_DELTA(np.array([0.05, 0.0, 0.0]))
- s --> APPLY_LIN_VEL_DELTA(np.array([-0.05, 0.0, 0.0]))
- q --> APPLY_LIN_VEL_DELTA(np.array([0.0, 0.05, 0.0]))
- e --> APPLY_LIN_VEL_DELTA(np.array([0.0, -0.05, 0.0]))
- a --> APPLY_ROT_VEL_DELTA(np.array([0.0, 0.0, 0.1]))
- d --> APPLY_ROT_VEL_DELTA(np.array([0.0, 0.0, -0.1]))
- up --> increment selected joint by 0.01 units
- down --> reduce reference joint position target for selected joint by 0.01 units
- right --> increment the id of the joint to be controlled with up/down keys
- left --> reduce the id of the joint to be controlled with up/down keys by 1
"""


def _return_arg1(x: Any, *args, **kwargs):  # pylint: disable=W0613
    return x


@dataclass
class GamePadButtonMappings:
    """By default all buttons do nothing to the GlobalMotionPlan object.

    Each button mapping can be replaced with another function callable
    that takes as argument a GlobalMotionPlan object and returns
    another GlobalMotionPlan object.
    """

    NORTH: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    SOUTH: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    EAST: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    WEST: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    DPAD_UP: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    DPAD_DOWN: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    DPAD_LEFT: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    DPAD_RIGHT: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    START_BUTTON: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    RB: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1
    LB: Callable[[GlobalMotionPlan], GlobalMotionPlan] = _return_arg1


@dataclass
class GamePadAnalogStickMappings:
    """By default all analog sticks do nothing to the GlobalMotionPlan object.

    Each analog stick mapping can be replaced with another function callable
    that takes as argument a GlobalMotionPlan object and a scaling value (float
    from the analog output (e.g. joystick value)), and returns a GlobalMotionPlan object.
    """

    LEFT_X: Callable[[GlobalMotionPlan, float], GlobalMotionPlan] = _return_arg1
    LEFT_Y: Callable[[GlobalMotionPlan, float], GlobalMotionPlan] = _return_arg1
    RIGHT_X: Callable[[GlobalMotionPlan, float], GlobalMotionPlan] = _return_arg1
    RIGHT_Y: Callable[[GlobalMotionPlan, float], GlobalMotionPlan] = _return_arg1
    RT: Callable[[GlobalMotionPlan, float], GlobalMotionPlan] = _return_arg1
    LT: Callable[[GlobalMotionPlan, float], GlobalMotionPlan] = _return_arg1


@dataclass
class GamePadMappings:
    button_mappings: GamePadButtonMappings = field(default_factory=GamePadButtonMappings)
    analog_mappings: GamePadAnalogStickMappings = field(default_factory=GamePadAnalogStickMappings)


def toggle_mapping(
    mapping: GamePadButtonMappings | GamePadAnalogStickMappings,
    key_name: str,
    bindings: Tuple[Callable[..., GlobalMotionPlan], Callable[..., GlobalMotionPlan]],
):
    """Toggle between two bindings (callables) for the specified attribute name in the mapping.

    Args:
        mapping (GamePadButtonMappings | GamePadAnalogStickMappings): The original mapping (should
            be mutable).
        key_name (str): Name of the attribute whose binding is to be modified.
        bindings (Tuple[Callable[..., GlobalMotionPlan], Callable[..., GlobalMotionPlan]]): The two
            bindings to switch between.
    """
    if getattr(mapping, key_name) == bindings[0]:
        setattr(mapping, key_name, bindings[1])
    else:
        setattr(mapping, key_name, bindings[0])


def _toggle_mapping_and_return_original_gplan(
    global_plan: GlobalMotionPlan,
    mapping: GamePadButtonMappings | GamePadAnalogStickMappings,
    key_name: str,
    bindings: Tuple[Callable[..., GlobalMotionPlan], Callable[..., GlobalMotionPlan]],
):
    toggle_mapping(mapping=mapping, key_name=key_name, bindings=bindings)
    return global_plan


_jpi2 = GlobalPlanJointPositionIncrementer()

DEFAULT_GAMEPAD_MAPPINGS = GamePadMappings(
    button_mappings=GamePadButtonMappings(
        START_BUTTON=Key2GlobalPlanHelper.TOGGLE_CUSTOM_IDLE,
        SOUTH=Key2GlobalPlanHelper.HOLD_POSITION,
        # DPAD_UP=Key2GlobalPlanHelper.APPLY_RELATIVE_POS_DELTA(
        #     position_delta=[0.0, 0.0, POS_DELTA]
        # ), # change base pose target height
        # DPAD_DOWN=Key2GlobalPlanHelper.APPLY_RELATIVE_POS_DELTA(
        #     position_delta=[0.0, 0.0, -POS_DELTA]
        # ), # change base pose target height
        DPAD_UP=partial(_jpi.increment_joint_target, position_delta=JOINT_POS_DELTA),
        DPAD_DOWN=partial(_jpi.increment_joint_target, position_delta=-JOINT_POS_DELTA),
        DPAD_LEFT=_jpi.decrease_joint_id,
        DPAD_RIGHT=_jpi.increase_joint_id,
    ),
    analog_mappings=GamePadAnalogStickMappings(
        LEFT_X=Analog2GlobalPlanHelper.APPLY_LIN_VEL_SCALED(
            max_linear_velocity=np.array([0.0, MAX_LIN_VEL, 0.0]),
            prev_vel_scaling=np.array([1.0, 0.0, 1.0]),
        ),
        LEFT_Y=Analog2GlobalPlanHelper.APPLY_LIN_VEL_SCALED(
            max_linear_velocity=np.array([MAX_LIN_VEL, 0.0, 0.0]),
            prev_vel_scaling=np.array([0.0, 1.0, 1.0]),
        ),
        RIGHT_X=Analog2GlobalPlanHelper.APPLY_ROT_VEL_SCALED(
            max_rotational_velocity=np.array([0.0, 0.0, MAX_ANG_VEL]),
            prev_vel_scaling=np.array([1.0, 1.0, 0.0]),
        ),
        RIGHT_Y=Analog2GlobalPlanHelper.APPLY_ROT_VEL_SCALED(
            max_rotational_velocity=np.array([0.0, MAX_ANG_VEL, 0.0]),
            prev_vel_scaling=np.array([1.0, 0.0, 1.0]),
        ),
    ),
)
"""Default joystick mapping when using JoystickInterface as global planner for sending
global plan messages.

NOTE: SOUTH, EAST, WEST, NORTH below indicates the keys on the right side of the joypad.
    (e.g. this is A, B, X, Y respectively on logitech joypad)

BUG: WEST and NORTH for logitech seems to be switched!!

- Start btn --> Toggle between CUSTOM and IDLE modes
- South --> HOLD_POSITION
- DPAD UP --> Increase joint position target for the current joint id by 0.01 units.
- DPAD DOWN --> Decrease joint position target for the current joint id by 0.01 units.
- DPAD LEFT --> Change the ID of the joint by -1 (affects actions of DPAD UP and DOWN buttons).
- DPAD LEFT --> Change the ID of the joint by +1 (affects actions of DPAD UP and DOWN buttons).
- LEFT Analog up/down --> apply linear velocity forward/backward (x axis) (max vel: 0.5 m/s)
    (OR) apply linear velocity up/down (z axis) (max vel: 0.5 m/s). Toggle behaviour of
    LEFT Analog up/down using `LB` button.
- LEFT Analog right/left --> apply linear velocity right/left (base frame) (max vel: 0.5 m/s)
- RIGHT Analog right/left --> apply rotational velocity (about z axis of base frame; yaw rate)
    (max vel: 1.0 rad/sec) (OR) apply rotational velocity (about x axis; roll rate)
    (max vel: 0.5 m/s). Toggle behaviour of RIGHT Analog right/left using `RB` button.
- RIGHT Analog up/down --> apply rotational velocity (about y axis of global frame; pitch rate)

- LB: This button changes the behaviour of the LEFT analog stick. Switches between sending
    linear velocity along X, and linear velocity along Z.
- RB: This button changes the behaviour of the RIGHT analog stick. Switches between sending
    angular velocity about X, and linear velocity about Z.
"""

# NOTE: Very ugly way of toggling what the analogs sticks do when RB, LB is pressed
DEFAULT_GAMEPAD_MAPPINGS.button_mappings.LB = partial(
    _toggle_mapping_and_return_original_gplan,
    mapping=DEFAULT_GAMEPAD_MAPPINGS.analog_mappings,
    key_name="LEFT_Y",
    bindings=[
        Analog2GlobalPlanHelper.APPLY_LIN_VEL_SCALED(
            max_linear_velocity=np.array([0.0, 0.0, MAX_LIN_VEL]),
            prev_vel_scaling=np.array([1.0, 1.0, 0.0]),
        ),
        Analog2GlobalPlanHelper.APPLY_LIN_VEL_SCALED(
            max_linear_velocity=np.array([MAX_LIN_VEL, 0.0, 0.0]),
            prev_vel_scaling=np.array([0.0, 1.0, 1.0]),
        ),
    ],
)
DEFAULT_GAMEPAD_MAPPINGS.button_mappings.RB = partial(
    _toggle_mapping_and_return_original_gplan,
    mapping=DEFAULT_GAMEPAD_MAPPINGS.analog_mappings,
    key_name="RIGHT_X",
    bindings=[
        Analog2GlobalPlanHelper.APPLY_ROT_VEL_SCALED(
            max_rotational_velocity=np.array([-MAX_ANG_VEL, 0.0, 0.0]),
            prev_vel_scaling=np.array([0.0, 1.0, 1.0]),
        ),
        Analog2GlobalPlanHelper.APPLY_ROT_VEL_SCALED(
            max_rotational_velocity=np.array([0.0, 0.0, MAX_ANG_VEL]),
            prev_vel_scaling=np.array([1.0, 1.0, 0.0]),
        ),
    ],
)



import copy
from typing import Tuple

from ...core.types import GlobalMotionPlan, RobotCmd, RobotState, LocalMotionPlan
from .agent_base import AgentBase
from ..local_planners import LocalPlannerBase
from ..controllers import ControllerBase


class PlannerControllerAgent(AgentBase):
    """A simple implementation of Agent using a local planner object and controller.

    This agent simply calls the local planner update and controller update steps in sequence.
    """

    def __init__(self, local_planner: LocalPlannerBase, controller: ControllerBase):
        """Constructor.

        A simple implementation of Agent using a local planner object and controller.
        This agent simply calls the local planner update and controller update steps in sequence.


        Args:
            local_planner (LocalPlannerBase): The local planner object to be used in the loop.
            controller (ControllerBase): The controller to be used.
        """
        self.local_planner = local_planner
        self.controller = controller

        self._latest_local_plan = None
        self._latest_ctrl_cmd = None

    def get_action(
        self,
        robot_state: RobotState,
        global_plan: GlobalMotionPlan,
        t: float = None,
        dt: float = None,
    ) -> RobotCmd:
        self._latest_local_plan = self.local_planner.generate_local_plan(
            robot_state=robot_state, global_plan=global_plan, t=t, dt=dt
        )
        self._latest_ctrl_cmd = self.controller.update(
            robot_state=robot_state,
            local_plan=self._latest_local_plan,
            t=t,
            dt=dt,
        )
        return self._latest_ctrl_cmd

    def get_latest_local_plan(self) -> LocalMotionPlan:
        return copy.deepcopy(self._latest_local_plan)

    def get_latest_ctrl_cmd(self) -> RobotCmd:
        return copy.deepcopy(self._latest_ctrl_cmd)

    def get_last_output(self) -> Tuple[LocalMotionPlan, RobotCmd]:
        return self.get_latest_local_plan(), self.get_latest_ctrl_cmd()

    def shutdown(self):
        self.controller.shutdown()
        self.local_planner.shutdown()

    def get_class_info(self) -> str:
        info = f"{self.get_class_name()}:\n"
        info += f"\t\t Controller: {self.controller.get_class_info()}\n"
        info += f"\t\t Local Planner: {self.local_planner.get_class_info()}"
        return info



from dataclasses import dataclass, field
import numpy as np
from typing import List, Tuple, Mapping, Any
import copy
from numbers import Number

from .tf_types import Pose3D, Twist, Vector3D
from ..logging import logging


@dataclass
class EndEffectorStates:
    """State estimate data relating to end-effectors of the robot."""

    ee_names: List[str] = None
    """Names of the end-effector frames."""
    ee_poses: List[Pose3D] = None
    """Pose of each end-effector.
    TODO: later, this may need to be of type: `List[FrameMotion]`.
    """
    ee_twists: List[Twist] = None
    """Twist for each end-effector"""
    contact_states: List[int] = None
    """Binary values denoting if end-effector(s) are in contact. 1 for contact, 0 for no-contact.
    Order is same as that of ee_names."""
    contact_forces: List[Vector3D] = None
    """A list of 3D vectors representing contact forces per end effector.

    Each row provides the x,y,z forces on one end_effector (order same as ee_names).
    """

    # internal mapping stored to make retrieval by name easier
    _ee_name_to_idx: Mapping[str, int] = field(default=None, init=False, repr=False)

    def get_state_of(
        self, ee_name: str
    ) -> Tuple[Pose3D | None, Twist | None, int | None, Vector3D | None]:
        """Get the end-effector states (pose, contact state, contact force) of the specified
        end-effector frame.

        This method is equivalent to directly using the `__getitem__` of this object. I.e.
        The two lines below are equivalent:
            >> p, v, c, f = ee_states.get_state_of(ee_name)
            >> p, v, c, f = ee_states[ee_name]

        Args:
            ee_name (str): The name of the end-effector frame.

        Raises:
            AttributeError: Raised if `ee_names` attribute was not defined for this object.
            KeyError: If the specified `ee_name` is not in the `ee_names` of this object.

        Returns:
            Tuple[Pose3D | None, Twist | None, int | None, Vector3D | None]: End-effector pose,
                twist, contact state (1 - contact, 0 - no contact), End-effector linear contact
                force (x,y,z).
        """
        if self._ee_name_to_idx is None:
            raise AttributeError(
                "`ee_names` attribute is not defined (is `None`). So this method cannot be used."
            )
        try:
            idx = self._ee_name_to_idx[ee_name]
        except KeyError as ke:
            raise KeyError(f"EE name '{ee_name}' not found in this object.") from ke
        else:
            ee_pose = None if self.ee_poses is None else self.ee_poses[idx]
            ee_twist = None if self.ee_twists is None else self.ee_twists[idx]
            contact_state = None if self.contact_states is None else self.contact_states[idx]
            contact_force = None if self.contact_forces is None else self.contact_forces[idx]
            return ee_pose, ee_twist, contact_state, contact_force

    def __getitem__(
        self, ee_name: str
    ) -> Tuple[Pose3D | None, Twist | None, int | None, Vector3D | None]:
        """Get the end-effector states (pose, contact state, contact force) of the specified
        end-effector frame.

        This method is equivalent to directly using the `get_state_of` method. I.e.
        The two lines below are equivalent:
            >> p, v, c, f = ee_states.get_state_of(ee_name)
            >> p, v, c, f = ee_states[ee_name]

        Args:
            ee_name (str): The name of the end-effector frame.

        Raises:
            AttributeError: Raised if `ee_names` attribute was not defined for this object.
            KeyError: If the specified `ee_name` is not in the `ee_names` of this object.

        Returns:
            Tuple[Pose3D | None, Twist | None, int | None, Vector3D | None]: End-effector pose,
                twist, contact state (1 - contact, 0 - no contact), End-effector linear contact
                force (x,y,z).
        """
        return self.get_state_of(ee_name=ee_name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "ee_names":
            super().__setattr__(
                "_ee_name_to_idx",
                (None if __value is None else {name: __value.index(name) for name in __value}),
            )
        elif __name in ["ee_poses", "ee_twists", "contact_states", "contact_forces"]:
            if __value is not None:
                if __name == "contact_states":
                    __value = np.array(__value)
                elif __name == "contact_forces":
                    __value = [np.array(v) if v is not None else None for v in __value]
        else:
            raise NameError(
                f"Tried to assign value to illegal parameter {__name} in object of type "
                f"{self.__class__.__name__}."
            )
        super().__setattr__(__name, __value)

    def extend(
        self,
        other: "EndEffectorStates",
        run_checks: bool = True,
        overwrite_existing: bool = False,
    ):
        """Extend the values of the attributes of this object using another instance of this class.

        Args:
            other (EndEffectorStates): The other instance whose values are to be used to extend
                this object.
            run_checks (bool, optional): If False, all values will be appended to existing ones
                for each attribute of this object, without checking for validity or repeats.
                Defaults to True.
            overwrite_existing (bool, optional): If set to True, will overwrite values with the
                new ones if same ee_name exists. Defaults to False.
        """
        if not (run_checks or overwrite_existing):
            self.ee_names = self.ee_names.copy() + other.ee_names
            self.ee_poses = self.ee_poses.copy() + other.ee_poses
            self.ee_twists = self.ee_twists.copy() + other.ee_twists
            self.contact_states = list(self.contact_states) + list(other.contact_states)
            self.contact_forces = list(self.contact_forces) + list(other.contact_forces)
        else:
            if not run_checks:
                logging.warning(
                    f"{self.__class__.__name__}: Setting `overwrite_existing` to True is"
                    " not allowed when `run_checks=False`. Will ignore the `run_checks`"
                    " flag for this now."
                )
            if not overwrite_existing:
                assert set(self.ee_names).isdisjoint(other.ee_names), (
                    f"{self.__class__.__name__}: EndEffectorStates object provided to extend this"
                    " class has ee names that are already present in current instance. To extend "
                    "this object by overwriting existing values for the common ee names, set "
                    "`override_estisting=True`"
                )
            ee_names = []
            ee_poses = []
            ee_twists = []
            ee_cs = []
            ee_forces = []
            for ee_name in self.ee_names + other.ee_names:
                if ee_name not in ee_names:
                    ee_names.append(ee_name)
                if ee_name in other.ee_names:
                    # if this ee is in the second object, info from that will be used,
                    # even if the same ee exists in the first object
                    p, v, c, f = other.get_state_of(ee_name=ee_name)
                else:
                    p, v, c, f = self.get_state_of(ee_name=ee_name)
                ee_poses.append(p)
                ee_twists.append(v)
                ee_cs.append(c)
                ee_forces.append(f)
            assert len(ee_names) == len(ee_poses) == len(ee_twists) == len(ee_cs)
            self.ee_names = ee_names
            self.ee_poses = ee_poses
            self.ee_twists = ee_twists
            self.contact_states = ee_cs
            self.contact_forces = ee_forces

    def update_from(self, other: "EndEffectorStates"):
        """Updates the attributes corresponding to each end-effector name of this object
        using values from another instance. Only the values of the end-effectors in the
        other instance will be modified; values of other end-effectors are left unchanged.

        Args:
            other (EndEffectorStates): The instance whose values will be used to modify
                the attributes of the current instance.
        """
        for n, j_name in enumerate(other.ee_names):
            if other.ee_poses is not None:
                self.ee_poses[self._ee_name_to_idx[j_name]] = other.ee_poses[n]
            if other.ee_twists is not None:
                self.ee_twists[self._ee_name_to_idx[j_name]] = other.ee_twists[n]
            if other.contact_states is not None:
                self.contact_states[self._ee_name_to_idx[j_name]] = other.contact_states[n]
            if other.contact_forces is not None:
                self.contact_forces[self._ee_name_to_idx[j_name]] = other.contact_forces[n]


@dataclass
class JointStates:
    """Joint state information/command/reference for an n-dof robot.
    All values are to be in the order specified in `joint_names`
    attribute."""

    joint_positions: np.ndarray = None
    joint_velocities: np.ndarray = None
    joint_efforts: np.ndarray = None
    joint_names: List[str] = None

    # internal mapping stored to make retrieval by name easier
    _joint_name_to_idx: Mapping[str, int] = field(default=None, init=False, repr=False)

    def get_state_of(self, joint_name: str) -> Tuple[float | None, float | None, float | None]:
        """Get the joint states (pos, vel, effort) of the specified joint.

        This method is equivalent to directly using the `__getitem__` of this object. I.e.
        The two lines below are equivalent:
            >> p, v, t = joint_states.get_state_of(joint_name)
            >> p, v, t = joint_states[joint_name]

        Args:
            joint_name (str): The name of the joint.

        Raises:
            AttributeError: Raised if `joint_names` attribute was not defined for this object.
            KeyError: If the specified `joint_name` is not in the `joint_names` of this object.

        Returns:
            Tuple[float | None, float | None, float | None]: pos, vel, effort of specified joint.
        """
        if self._joint_name_to_idx is None:
            raise AttributeError(
                "`joint_names` attribute is not defined (is `None`). So this method cannot be used."
            )
        try:
            idx = self._joint_name_to_idx[joint_name]
        except KeyError as ke:
            raise KeyError(f"Joint '{joint_name}' not found in this object.") from ke
        else:
            joint_position = None if self.joint_positions is None else self.joint_positions[idx]
            joint_velocity = None if self.joint_velocities is None else self.joint_velocities[idx]
            joint_effort = None if self.joint_efforts is None else self.joint_efforts[idx]
            return joint_position, joint_velocity, joint_effort

    def __getitem__(self, joint_name: str) -> Tuple[float | None, float | None, float | None]:
        """Get the joint states (pos, vel, effort) of the specified joint.

        This method is equivalent to directly using the `get_state_of` method of this class. I.e.
        The two lines below are equivalent:
            >> p, v, t = joint_states.get_state_of(joint_name)
            >> p, v, t = joint_states[joint_name]

        Args:
            joint_name (str): The name of the joint.

        Raises:
            AttributeError: Raised if `joint_names` attribute was not defined for this object.
            KeyError: If the specified `joint_name` is not in the `joint_names` of this object.

        Returns:
            Tuple[float | None, float | None, float | None]: pos, vel, effort of specified joint.
        """
        return self.get_state_of(joint_name=joint_name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name == "joint_names":
            super().__setattr__(
                "_joint_name_to_idx",
                (None if __value is None else {name: __value.index(name) for name in __value}),
            )
        elif __name in ["joint_positions", "joint_velocities", "joint_efforts"]:
            if __value is not None:
                __value = np.array(__value)
        else:
            raise NameError(
                f"Tried to assign value to illegal parameter {__name} in object ot type "
                f"{self.__class__.__name__}."
            )
        super().__setattr__(__name, __value)

    def extend(
        self,
        other: "JointStates",
        run_checks: bool = True,
        overwrite_existing: bool = False,
    ):
        """Extend the values of the attributes of this object using another instance of this class.

        Args:
            other (JointStates): The other instance whose values are to be used to extend
                this object.
            run_checks (bool, optional): If False, all values will be appended to existing ones
                for each attribute of this object, without checking for validity or repeats.
                Defaults to True.
            overwrite_existing (bool, optional): If set to True, will overwrite values with the
                new ones if same joint_name exists. Defaults to False.
        """
        if not (run_checks or overwrite_existing):
            self.joint_names = self.joint_names.copy() + other.joint_names
            self.joint_positions = np.append(self.joint_positions, other.joint_positions)
            self.joint_velocities = np.append(self.joint_velocities, other.joint_velocities)
            self.joint_efforts = np.append(self.joint_efforts, other.joint_efforts)
        else:
            if not run_checks:
                logging.warning(
                    f"{self.__class__.__name__}: Setting `overwrite_existing` to True is"
                    " not allowed when `run_checks=False`. Will ignore the `run_checks`"
                    " flag for this now."
                )
            if not overwrite_existing:
                assert set(self.joint_names).isdisjoint(other.joint_names), (
                    f"{self.__class__.__name__}: JointState object provided for extending this"
                    " class has joint names that are already present in current instance. To extend "
                    "this object by overwriting existing values for the common joints, set "
                    "`override_estisting=True`"
                )
            j_names = []
            j_pos = []
            j_vel = []
            j_eff = []
            for j_name in self.joint_names + other.joint_names:
                if j_name not in j_names:
                    j_names.append(j_name)
                if j_name in other.joint_names:
                    # if this joint is in the second object, info from that will be used,
                    # even if the same joint exists in the first object
                    p, v, t = other.get_state_of(joint_name=j_name)
                else:
                    p, v, t = self.get_state_of(joint_name=j_name)
                j_pos.append(p)
                j_vel.append(v)
                j_eff.append(t)
            assert len(j_names) == len(j_pos) == len(j_vel) == len(j_eff)
            self.joint_names = j_names
            self.joint_positions = j_pos
            self.joint_velocities = j_vel
            self.joint_efforts = j_eff

    def update_from(self, other: "JointStates"):
        """Updates the attributes corresponding to each joint name of this object
        using values from another instance. Only the values of the joints in the
        other instance will be modified; values of other joints are left unchanged.

        Args:
            other (JointStates): The instance whose values will be used to modify
                the attributes of the current instance.
        """
        for n, j_name in enumerate(other.joint_names):
            if other.joint_positions is not None:
                self.joint_positions[self._joint_name_to_idx[j_name]] = other.joint_positions[n]
            if other.joint_velocities is not None:
                self.joint_velocities[self._joint_name_to_idx[j_name]] = other.joint_velocities[n]
            if other.joint_efforts is not None:
                self.joint_efforts[self._joint_name_to_idx[j_name]] = other.joint_efforts[n]


@dataclass
class StateEstimates:
    """Possible outputs of a robot state estimator.

    This should contain state estimates that are computed and not typically observed
    directly using sensors.

    NOTE: It might be worth making base state estimates into type FrameMotion.
    """

    pose: Pose3D = field(default_factory=Pose3D)
    """Pose of the robot's base link in the world frame."""
    twist: Twist = field(default_factory=Twist)
    """Velocity of the robot in the base frame; linear and angular."""
    end_effector_states: EndEffectorStates = field(default_factory=EndEffectorStates)
    """State estimate data relating to end-effectors of this robot."""

    def extend(
        self,
        other: "StateEstimates",
        run_checks: bool = True,
        overwrite_existing: bool = False,
    ):
        """Extend the values of the `end_effector_states` attribute of this object using another
        instance of this class.

        Args:
            other (StateEstimates): The other instance whose values are to be used to extend
                this object.
            run_checks (bool, optional): If False, all values will be appended to existing ones
                for each attribute of `joint_states`, without checking for validity or repeats.
                Defaults to True.
            overwrite_existing (bool, optional): If set to True, will overwrite values with the
                new ones in `joint_states` if same joint_name exists. Defaults to False.
        """
        self.end_effector_states.extend(
            other=other.end_effector_states,
            run_checks=run_checks,
            overwrite_existing=overwrite_existing,
        )

    def update_from(self, other: "StateEstimates", update_base_state: bool):
        """Updates the attributes corresponding to each joint name of this object
        using values from another instance. Only the values of the joints in the
        other instance will be modified; values of other joints are left unchanged.

        Args:
            other (StateEstimates): The instance whose values will be used to modify
                the attributes of the current instance.
            update_base_state (bool): If set to True, will also modify the `pose` and
                `twist` attributes from the other instance.
        """
        if update_base_state:
            self.pose = other.pose
            self.twist = other.twist
        self.end_effector_states.update_from(other=other.end_effector_states)


@dataclass
class RobotState:
    """Represents the state of a robot.

    All joint values should be in the order mentioned in joint_names.
    """

    state_estimates: StateEstimates = field(default_factory=StateEstimates)
    """State estimator output."""
    joint_states: JointStates = field(default_factory=JointStates)
    """Proprioceptive data from the robot sensors/interfaces."""

    def extend(
        self,
        other: "RobotState",
        run_checks: bool = True,
        overwrite_existing: bool = False,
    ):
        """Extend the values of the `state_estimates` and `joint_states` attribute of this object
        using another instance of this class.

        Args:
            other (RobotState): The other instance whose values are to be used to extend
                this object.
            run_checks (bool, optional): If False, all values will be appended to existing ones
                for each attribute of `joint_states` and `state_estimates`, without checking for
                validity or repeats. Defaults to True.
            overwrite_existing (bool, optional): If set to True, will overwrite values with the
                new ones if same joint name of end-effector name exists in the other instance.
                Defaults to False.
        """
        self.state_estimates.extend(
            other=other.state_estimates,
            run_checks=run_checks,
            overwrite_existing=overwrite_existing,
        )
        self.joint_states.extend(
            other=other.joint_states,
            run_checks=run_checks,
            overwrite_existing=overwrite_existing,
        )

    def update_from(self, other: "RobotState", update_base_state: bool):
        """Updates the attributes corresponding to each joint name and end-effector name
        of this object using values from another instance. Only the values of the joints in the
        other instance will be modified; values of other joints are left unchanged.


        Args:
            other (RobotState): The instance whose values are to be used for modifying this
                object.
            update_base_state (bool): If set to True, will also modify the `pose` and
                `twist` attributes from the other instance's state_estimates.
        """
        self.state_estimates.update_from(
            other=other.state_estimates, update_base_state=update_base_state
        )
        self.joint_states.update_from(other=other.joint_states)


# pylint: disable=W0212
@dataclass
class RobotCmd:
    """Object containing low-level commands to be sent directly to the robot.

    This is computed by the controller. All joint orders should be same as the order in RobotState.

    This object follows the PVT-PD (position, velocity, effort, stiffness (P), damping (D))
    actuation command protocol (but for sequence of actuators, instead of just one).

    Attributes (all values have to be filled in the order defined in `joint_commands.joint_names`):
        joint_commands.joint_names (List[str])
        joint_commands.joint_positions (np.ndarray)
        joint_commands.joint_velocities (np.ndarray)
        joint_commands.joint_efforts (np.ndarray)
        Kp (np.ndarray)
        Kd (np.ndarray)
    """

    joint_commands: JointStates = field(default_factory=JointStates)
    Kp: np.ndarray = None
    """Array of joint position gains for each joint (stiffness) in `joint_commands.joint_names`
    order."""
    Kd: np.ndarray = None
    """Array of joint velocity gains (or damping if `joint_velocities=None` or zeros) in 
    `joint_commands.joint_names` order."""

    def set_joint_gains(self, joint_names: List[str], Kp: np.ndarray = None, Kd: np.ndarray = None):
        """Set Kp and/or Kd gains for a list of specified joints.

        Args:
            joint_names (List[str]): Names of joints whose gains are to be modified.
            Kp (np.ndarray, optional): New Kp value (per specified joint). Defaults to None (no
                change).
            Kd (np.ndarray, optional): New Kd value (per specified joint). Defaults to None (no
                change).
        """
        assert (
            self.joint_commands.joint_names is not None
        ), "Joint names not provided in `joint_commands.joint_names`."
        if self.Kp is None:
            self.Kp = np.zeros(len(self.joint_commands.joint_names))
        if self.Kd is None:
            self.Kd = np.zeros(len(self.joint_commands.joint_names))
        for n, jname in enumerate(joint_names):
            if Kp is not None:
                self.Kp[self.joint_commands._joint_name_to_idx[jname]] = Kp[n]
            if Kd is not None:
                self.Kd[self.joint_commands._joint_name_to_idx[jname]] = Kd[n]

    def get_command_for_joint(self, joint_name: str) -> Tuple[float, float, float, float, float]:
        """Get the commands for the specified joint.

        Args:
            joint_name (str): Name of the joint.

        Returns:
            Tuple[float, float, float, float, float]: Kp, Kd, position, velocity, effort
        """
        jid = self.joint_commands._joint_name_to_idx[joint_name]
        return (
            self.Kp[jid],
            self.Kd[jid],
            self.joint_commands.joint_positions[jid],
            self.joint_commands.joint_velocities[jid],
            self.joint_commands.joint_efforts[jid],
        )

    @classmethod
    def createZeros(cls: "RobotCmd", dof: int, joint_names: List[str] = None) -> "RobotCmd":
        """Static class method to create a pre-populated RobotCmd with `dof` number of joints.

        All values are set to zeros (including gains).

        Args:
            dof (int): Number of joints. Values will be set to numpy arrays of this size.
            joint_names (List[float], optional): List of joint names. By default, this is None,
                and will have to be filled later if needed.

        Returns:
            RobotCmd: An instance of RobotCmd with all values zero with the specified number of
                joints.
        """
        assert isinstance(dof, int)
        if joint_names is not None:
            assert len(joint_names) == dof
        return cls(
            joint_commands=JointStates(
                joint_positions=np.zeros(dof),
                joint_velocities=np.zeros(dof),
                joint_efforts=np.zeros(dof),
                joint_names=copy.deepcopy(joint_names),
            ),
            Kp=np.zeros(dof),
            Kd=np.zeros(dof),
        )

    @classmethod
    def fromJointStates(
        cls: "RobotCmd",
        joint_states: JointStates,
        Kp: float | np.ndarray,
        Kd: float | np.ndarray,
    ) -> "RobotCmd":
        """Static class method to create a RobotCmd object using a provided JointStates object.

        This method will directly use values from the provided JointStates object and set them
        as the values for `joint_commands`. Kp and Kd are read from the arguments.

        Args:
            joint_states (JointStates): The JointStates object whose values are to be used.
            Kp (float | np.ndarray): Kp for position tracking.
            Kd (float | np.ndarray): Kd for velocity tracking (or damping)

        Returns:
            RobotCmd: An instance of RobotCmd with all values from the provided JointStates object.
        """
        return cls(joint_commands=copy.deepcopy(joint_states), Kp=Kp, Kd=Kd)

    def __setattr__(self, prop, val):
        # custom handling to make sure kp and kd dimensions and signs are valid
        if prop in ["Kp", "Kd"] and val is not None:
            if prop == "Kp":
                _compare_with = self.joint_commands.joint_positions
            else:
                _compare_with = self.joint_commands.joint_velocities
            if _compare_with is not None:
                if isinstance(val, Number):
                    val = [val] * len(_compare_with)
                else:
                    if len(val) == 1:
                        val = [val[0]] * len(_compare_with)
                    assert len(val) == len(
                        _compare_with
                    ), f"{prop} dimensions ({len(val)}) should match dimension of {'joint_positions' if prop == 'Kp' else 'joint_velocities'} ({len(_compare_with)}) in robot cmd."
            assert np.all(
                val == np.abs(val)
            ), f"All values for {prop} should be positive. Invalid value provided: {val}"
            val = np.array(val)
        super().__setattr__(prop, val)

    def extend(
        self,
        other: "RobotCmd",
        run_checks: bool = True,
        overwrite_existing: bool = False,
    ):
        """Extend the values of the attributes of this object using another instance of this class.

        Args:
            other (RobotCmd): The other instance whose values are to be used to extend
                this object.
            run_checks (bool, optional): If False, all values will be appended to existing ones
                for each attribute of this object, without checking for validity or repeats.
                Defaults to True.
            overwrite_existing (bool, optional): If set to True, will overwrite values with the
                new ones if same joint_name exists. Defaults to False.
        """
        new_kps = np.append(self.Kp, other.Kp)
        new_kds = np.append(self.Kd, other.Kd)
        j_names = self.joint_commands.joint_names + other.joint_commands.joint_names
        self.joint_commands.extend(
            other=other.joint_commands,
            run_checks=run_checks,
            overwrite_existing=overwrite_existing,
        )
        if not (run_checks or overwrite_existing):
            self.Kp = new_kps
            self.Kd = new_kds
        else:
            self.set_joint_gains(joint_names=j_names, Kp=new_kps, Kd=new_kds)

    def update_from(self, other: "RobotCmd"):
        """Updates the attributes corresponding to each joint name of this object
        using values from another instance. Only the values of the joints in the
        other instance will be modified; values of joints not in the other instance
        are left unchanged.

        Args:
            other (RobotCmd): The instance whose values will be used to modify
                the attributes of the current instance.
        """
        for n, jname in enumerate(other.joint_commands.joint_names):
            self.Kp[self.joint_commands._joint_name_to_idx[jname]] = other.Kp[n]
            self.Kd[self.joint_commands._joint_name_to_idx[jname]] = other.Kd[n]
            # more efficient to update all joint states and gains here in the same loop than
            # to call `self.joint_commands.update_from(other=other.joint_commands)`
            if other.joint_commands.joint_positions is not None:
                self.joint_commands.joint_positions[
                    self.joint_commands._joint_name_to_idx[jname]
                ] = other.joint_commands.joint_positions[n]
            if other.joint_commands.joint_velocities is not None:
                self.joint_commands.joint_velocities[
                    self.joint_commands._joint_name_to_idx[jname]
                ] = other.joint_commands.joint_velocities[n]
            if other.joint_commands.joint_efforts is not None:
                self.joint_commands.joint_efforts[self.joint_commands._joint_name_to_idx[jname]] = (
                    other.joint_commands.joint_efforts[n]
                )



class PyRCFExceptionBase(Exception): ...


# CtrlLoop Signal
class CtrlLoopExitSignal(KeyboardInterrupt):
    """Request to exit the control loop."""


# UI Errors
class UIException(IOError):
    """Exceptions relating to user interfaces (UIBase objects)."""


class NotConnectedError(UIException):
    """Exception to be thrown when a UI interface could not be detected or was disconnected."""



"""Utilities for timer and rate-related functionalities."""

from time import perf_counter, sleep, time
from abc import ABC, abstractmethod
from ..core.logging import logging


class ClockBase(ABC):
    """Base class for any Clock implementation for pyrcf control loop and components."""

    @abstractmethod
    def get_time(self) -> float:
        """Get the latest time from this clock.

        Raises:
            NotImplementedError: Raised if this method is not implemented by the child class.

        Returns:
            float: the current time in seconds.
        """
        raise NotImplementedError("This method has to be implemented in the child class")


class PythonEpochClock(ClockBase):
    """ClockBase implementation to query system time using `time.time()`."""

    def get_time(self) -> float:
        return time()


class PythonPerfClock(ClockBase):
    """ClockBase implementation to query counter using `time.perf_counter()`. Only
    meaningful when comparing values from other `time.perf_counter()` calls."""

    def get_time(self) -> float:
        return perf_counter()


class RateLimiter:
    """
    Modified from https://github.com/upkie/loop-rate-limiters/blob/main/loop_rate_limiters/rate_limiter.py.

    Original License terms below:

    Regulate the frequency between calls to the same instruction.

    This rate limniter is meant to be used in e.g. a loop or callback function.
    It is, in essence, the same as rospy.Rate_. It assumes Python's performance
    counter never jumps backward nor forward, so that it does not handle such
    cases contrary to rospy.Rate_.

    .. _rospy.Rate:
        https://github.com/ros/ros_comm/blob/noetic-devel/clients/rospy/src/rospy/timer.py

    Attributes:
        name: Human-readable name used for logging.
        warn: If set (default), warn when the time between two calls
            exceeded the rate clock.

    #
    # Copyright 2022 Stéphane Caron
    # Copyright 2023 Inria
    #
    # Licensed under the Apache License, Version 2.0 (the "License");
    # you may not use this file except in compliance with the License.
    # You may obtain a copy of the License at
    #
    #     http://www.apache.org/licenses/LICENSE-2.0
    #
    # Unless required by applicable law or agreed to in writing, software
    # distributed under the License is distributed on an "AS IS" BASIS,
    # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    # See the License for the specific language governing permissions and
    # limitations under the License.
    """

    __period: float
    __slack: float
    __next_tick: float
    __freq: float
    name: str
    warn: bool

    def __init__(
        self,
        frequency: float,
        name: str = "rate limiter",
        warn: bool = False,
        clock: ClockBase = PythonPerfClock(),
    ):
        """Initialize rate limiter.

        Args:
            frequency: Desired frequency in hertz.
            name: Human-readable name used for logging.
            warn: If set (default), warn when the time between two calls
                exceeded the rate clock.
        """
        self.set_frequency(frequency=frequency)
        self.change_clock(clock=clock)
        self.name = name
        self.warn = warn

    def set_frequency(self, frequency: float):
        self.__period = 1.0 / frequency
        self.__freq = frequency

    def change_clock(self, clock: ClockBase):
        self.__clock = clock
        self.__next_tick = self.__clock.get_time() + self.__period
        self.__slack = 0.0

    @property
    def clock(self) -> ClockBase:
        """The clock used by this class.

        Use `get_time()` on this attribute to get current time in seconds.

        Returns:
            ClockBase: The clock object used by this class.
        """
        return self.__clock

    @property
    def dt(self) -> float:
        """Desired period between two calls to :func:`sleep`, in seconds."""
        return self.__period

    @property
    def next_tick(self) -> float:
        """Time of next clock tick."""
        return self.__next_tick

    @property
    def period(self) -> float:
        """Desired period between two calls to :func:`sleep`, in seconds."""
        return self.__period

    @property
    def frequency(self) -> float:
        """Desired frequency between two calls to :func:`sleep`, in seconds."""
        return self.__freq

    @property
    def slack(self) -> float:
        """Slack duration computed at the last call to :func:`sleep`.

        This duration is in seconds.
        """
        return self.__slack

    def remaining(self) -> float:
        """Get the time remaining until the next expected clock tick.

        Returns:
            Time remaining, in seconds, until the next expected clock tick.
        """
        return self.__next_tick - self.__clock.get_time()

    def sleep(self):
        """Sleep for the duration required to regulate inter-call frequency."""
        self.__slack = self.__next_tick - self.__clock.get_time()
        if self.__slack > 0.0:
            sleep(self.__slack)
        elif self.warn and self.__slack < -0.1 * self.period:
            logging.warning(
                "%s is late by %f [ms]",
                self.name,
                round(1e3 * self.__slack, 1),
            )
        self.__next_tick = self.__clock.get_time() + self.__period


class RateTrigger:
    """
    Can be used to execute something at a specific rate. Checking
    the `.triggered()` method of this class will return True or
    False depending on whether the specified rate has reached.

    NOTE: Warning: When using this with RateLimiter in the same
    loop, try to keep the rate of this class a factor of the rate
    of the RateLimiter object. Otherwise, the RateTrigger will
    not be able to keep the required rate (because RateLimiter
    enforces a sleep). Make sure to use same clock as well.
    """

    def __init__(self, rate: float, clock: ClockBase = PythonPerfClock()):
        """RateTrigger Constructor.

        Can be used to execute something at a specific rate. Checking
        the `.triggered()` method of this class will return True or
        False depending on whether the specified rate has reached.

        NOTE: Warning: When using this with RateLimiter in the same
        loop, try to keep the rate of this class a factor of the rate
        of the RateLimiter object. Otherwise, the RateTrigger will
        not be able to keep the required rate (because RateLimiter
        enforces a sleep).

        Args:
            rate (float): The desired frequency at which this should
                trigger True.
        """
        self.__period = 1.0 / rate
        self.__next_tick = None
        self.change_clock(clock=clock)

    def change_clock(self, clock: ClockBase):
        self.__clock = clock
        self.__next_tick = self.__clock.get_time() + self.__period

    @property
    def clock(self) -> ClockBase:
        """The clock used by this class.

        Use `get_time()` on this attribute to get current time in seconds.

        Returns:
            ClockBase: The clock object used by this class.
        """
        return self.__clock

    @property
    def dt(self) -> float:
        """Desired period between two calls to :func:`sleep`, in seconds."""
        return self.__period

    @property
    def next_tick(self) -> float:
        """Time of next clock tick."""
        return self.__next_tick

    @property
    def period(self) -> float:
        """Desired period between two calls to :func:`sleep`, in seconds."""
        return self.__period

    def triggered(self) -> bool:
        """Check if the specified rate has been triggered."""
        if self.__next_tick - self.__clock.get_time() <= 0.0:
            self.__next_tick = self.__clock.get_time() + self.__period
            return True

        return False

    @property
    def has_triggered(self) -> bool:
        """Attribute to check if the specified rate has been triggered."""
        return self.triggered()

    def __call__(self) -> bool:
        """Calling this class object directly will check if the specified rate has been
        triggered."""
        return self.triggered()
